/*  PRNCOM  */
/* Data I/O via Paralel (Centronics) printer port of the (IBM) PC-AT/XT ;
 * Description of configuration, control and use ;
 */
/*  by : Erik Margan  */

/*  */
/* Port Address Map ( 16-Bit addr. ) supported by MS-DOS :
 *   Ŀ
 *    port  DATA (base addr.)  STATUS   CONTROL 
 *   Ĵ
 *    LPT1       0x378         0x379     0x37A  
 *   Ĵ
 *    LPT2       0x278         0x279     0x27A  
 *   Ĵ
 *    LPT3       0x3BC         0x3BD     0x3BE  
 *   
 *    !!!! NOTE : On some PC's LPT1 is at 0x3BC !!!!
 */
/*  */
/* The 25-pin connector ( input/output definition on the PC-side ) :
 *
 *            <--  _Strobe    1 
 *                                  14    _AUTO-FD-XT    -->
 *            <--       D0    2 
 *                                  15    Error         <--
 *            <--       D1    3 
 *                                  16    INIT           -->
 *            <--       D2    4 
 *                                  17    _SLCT-IN       -->
 *            <--       D3    5 
 *                                  18    GND
 *            <--       D4    6 
 *                                  19    GND
 *            <--       D5    7 
 *                                  20    GND
 *            <--       D6    8 
 *                                  21    GND
 *            <--       D7    9 
 *                                  22    GND
 *             -->     ACK   10 
 *                                  23    GND
 *             -->   _BUSY   11 
 *                                  24    GND
 *             -->      PE   12 
 *                                  25    GND
 *             -->    SLCT   13 
 */
/*  */
/* REGISTER FUNCTIONS :
 * DATA    is the data-output port. The bits D0-D7 apear as latched TTL-level
 *         at the pins 2-9 of the 25-pin connector. The port can be read, but
 *         the tri-state output enable is fixed so the port can not be used
 *         bidirectionally. A series resistor and a capacitor to GND reduce
 *         the cross-talk in the cable lines.
 *
 * STATUS  is the status input port. Not all 8 bits are used :
 *   Ŀ
 *    bit      Function          pin    
 *   Ĵ
 *    0,1,2    Not used          -      
 *   Ĵ
 *     3       Error             15     
 *   Ĵ
 *     4       SLCT (on-line)    13     
 *   Ĵ
 *     5       PE (paper out)    12     
 *   Ĵ
 *     6       ACK (IRQ)         10     
 *   Ĵ
 *     7       _BUSY             11     
 *   
 *    Bit 7 is inverted from the connector to the bus.
 *    Bit 6 can be used as an external interrupt (IRQ7 on LPT1).
 *
 *
 * CONTROL provides the output control signals. Not all 8 bits are used.
 *         Bits 0, 1 and 3 are inverted as they appear on the connector.
 *         Only the _STROBE signal has an output filter. As the DATA, the
 *         outputs can be read; the bit inversion is corrected when reading.
 *   Ŀ
 *    bit      Function        pin  
 *   Ĵ     * * * * * * * * * * * * * * *
 *     0       _STROBE          1        *   A 4k7 pull-up resistor  *
 *   Ĵ     * is recomended at the re-  *
 *     1       _AUTO-FD-XT      14       * ceiving side to reduce    *
 *   Ĵ     * the delay on the rising   *
 *     2       INIT             16       * edge thus speeding data   *
 *   Ĵ     * transfer.                 *
 *     3       _SLCT-IN         17       * * * * * * * * * * * * * * *
 *   Ĵ
 *     4       IRQ-EN (IRQ7)    -   
 *   Ĵ
 *    5,6,7    Not used         -   
 *   
 */
/*  Port register addresses are used as global variables  */

int pdata ;         /* LPT1 = 0x378 , LPT2 = 0x278 , LPT3 = 0x3BC */
int pstat ;         /* pstat = pport + 1 ; */
int pctrl ;         /* pctrl = pport + 2 ; */

#define IDLE                0x00
#define READ_STATUS         0x00

#define WRITE_ADDR_B        0x00

#define STROBE              0x01

#define MASTER_RESET        0x02

#define READ_DATA_HI        0x02
#define READ_DATA_LO        0x06

#define WRITE_DATA_A        0x04
#define LATCH_DATA_A		0x06


#define NOOP                0x08

/*  */
