/*  Patch-Panel 1(B/F/D) PCB Test Program ( PTX9.C )  */
/*												Erik Margan, 20021129       */

/*  Standard QC include headers  */
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <conio.h>
#include <bios.h>
#include <dos.h>
#include <time.h>
#include <graph.h>
#include <math.h>

/*  PRN-Port Interface include header  */
#include "prncom.h"

/*  Global definitions of constant  */
#define MAXSTRING 80
#define PAGESIZE 15
#define MAXSOCKNUM 9

/*  Global variables  */

int pdatareg, pctrlreg, pstatreg ;
/* pstatreg = pdatareg + 1 ; */
/* pctrlreg = pdatareg + 2 ; */
/* pdatareg address can be : LPT1 = 0x378 , LPT2 = 0x278 , LPT3 = 0x3BC */

int flag_test = 0 ;		/* program testing mode flag, active if 1 */
int flag_speed = 0 ;	/* PRN R/W op. can be slowed if PC speed is 2 high */

int lastpp1 = 0 ;		/* serial number of the last board tested */
char flag_type = 'X' ;  /* PCB type B = barrel, F = forward, D = TPPF1_D9 */

						/* text buffer for message display */
char txtbuf[ MAXSTRING+1 ] ;
						/* file list, 13 characters wide */
char filelist[ PAGESIZE ][ 13 ] ;


/*  */
void ClearKbBuf( void )
{
while( _bios_keybrd( _KEYBRD_READY ) )
	_bios_keybrd( _KEYBRD_READ ) ;
}


/*  */
int WaitKey( int wait_flag )
{
int key ;

if ( wait_flag )
	{
	ClearKbBuf() ;

	while( !( _bios_keybrd( _KEYBRD_READY ) ) )
		{ ; }
	}

if( _bios_keybrd( _KEYBRD_READY ) )
	key = _bios_keybrd( _KEYBRD_READ ) ;

if ( wait_flag )
	ClearKbBuf() ;

return key ;
}


/*  */
void ScrLine( int row, int col )
{
_settextposition( row, col ) ;
_outtext( txtbuf ) ;
}


/*  */
void ClearLines( int startline, int endline )
{
int kx ;

for ( kx = startline ; kx <= endline ; kx++ )
	{
	sprintf( txtbuf, "                                                       " ) ;
	ScrLine( kx, 1 ) ;
	}
}


/*  */
void ErrorLine( int bp, int wk )
{
_settextposition( 24, 1 ) ;
_outtext( txtbuf ) ;

if ( bp )
	printf( "\a" ) ;

if ( wk )
	{
	_settextposition( 25, 5 ) ;
	_outtext( "Press any key ... " ) ;
	WaitKey( 1 ) ;
	ClearLines( 24, 25 ) ;
	}
}


/*  */
/* Function :		Delay55ms()
 * Purpose :		Wait for a specified time, checking keyboard.
 * Parameters :		(in) nx55 ..... integer >= 0 by which the basic
 *									55ms delay is multiplied
 * Variables :		t0 ............ returned time value
 *					flag_kb ....... keyboard flag ; break if > 0
 * Return :			flag_kb
 */

int Delay55ms( int nx55 )
{
clock_t t0 ;
int flag_kb ;

flag_kb = 0 ;

if ( nx55 < 0 )
	nx55 = 0 ;

do	{
	t0 = clock() ;
	while( clock() == t0 )
		{ ; }
	} while( nx55-- && !kbhit() ) ;

if ( kbhit() )
	flag_kb = 1 ;

return flag_kb ;
}


/*  */
/* Function :		CheckTime()
 * Purpose :		Check if the specified time has expired.
 * Parameters :		(in) nsec ..... integer > 0, num of seconds to check.
 *					(i/o) tzero ... clock_t type, start time if non-zero,
 *									get start time if zero.
 * Variables :		tx ............ float type, containing the actual time.
 *					t_flag ........ delay termination test flag.
 *									a key was pressed during the delay.
 * Return :			t_flag ........ < 0 if no key and time not expired;
 *									= 0 if no key and time expired;
 *									> 0 if key interrupt request.
 */

int CheckTime( float *delta_t, float *zero_t )
{
int t_flag, flag_kb ;

float dt, zt ;

/* char txtbuf[20] ; */

t_flag = 0 ;

zt = *zero_t ;
dt = *delta_t ;

if ( zt == 0.0 )
	{
	zt = (float)( clock() + dt * CLOCKS_PER_SEC ) ;
	}
else
	{
	dt = ( zt - (float)clock() ) / CLOCKS_PER_SEC ;
	if ( dt < 0.1 )
		t_flag = 1 ;
	}

sprintf( txtbuf, "Time = % 5.1f s  ", dt ) ;
ScrLine( 6, 35 ) ;

*delta_t = dt ;
*zero_t = zt ;

return t_flag ;
}


/*  */
/* Function :         InitPrnPort()
 * Purpose :          Initialize PRN port for bus control
 *                    using a virtual handshaking protocol.
 * Parameters :		  (none)
 * Variables :        state     port_C status
 *                    hwerr     hardware error flag, 0 = no error
 * Return :           hwerr
 */

int InitPrnPort( void )
{
int state, hwerr ;
/* char txtbuf[80] ; */

hwerr = 0 ;

/* pdatareg = 0x378 ; */ /* test lpt3 first ! */
pdatareg = 0x3BC ;

outp( pdatareg, 0x55 ) ;
state = inp( pdatareg ) ;
if ( state != 0x55 )
    hwerr |= 1 ;

outp( pdatareg, 0xAA ) ;
state = inp( pdatareg ) ;
if ( state != 0xAA )
    hwerr |= 2 ;

if ( hwerr )
    {
    hwerr = 0 ;
	/* pdatareg = 0x3BC ; */
    pdatareg = 0x378 ;

    outp( pdatareg, 0x55 ) ;
    state = inp( pdatareg ) ;
    if ( state != 0x55 )
        hwerr |= 1 ;

    outp( pdatareg, 0xAA ) ;
    state = inp( pdatareg ) ;
    if ( state != 0xAA )
        hwerr |= 2 ;
    }

if ( hwerr )
    {
    hwerr = 0 ;
    pdatareg = 0x278 ;

    outp( pdatareg, 0x55 ) ;
    state = inp( pdatareg ) ;
    if ( state != 0x55 )
        hwerr |= 1 ;

    outp( pdatareg, 0xAA ) ;
    state = inp( pdatareg ) ;
    if ( state != 0xAA )
        hwerr |= 2 ;
    }

if ( hwerr )
    {
    hwerr = 0 ;
    pdatareg = 0x378 ;
	sprintf( txtbuf, ">> WARNING : PRN-Port error ! Running at 0x378 !" ) ;
	ErrorLine( 1, 1 ) ;
    }

pstatreg = pdatareg + 1 ;
pctrlreg = pdatareg + 2 ;

outp( pdatareg, 0x00 ) ;

/* outp( pctrlreg, NOOP ) ; */
outp( pctrlreg, MASTER_RESET ) ;
inp( pstatreg ) ;
outp( pctrlreg, MASTER_RESET | STROBE ) ;
inp( pstatreg ) ;
outp( pctrlreg, MASTER_RESET ) ;
inp( pstatreg ) ;
outp( pctrlreg, IDLE ) ;
inp( pstatreg ) ;

state = inp( pstatreg ) & 0xF8 ;
if ( state == 0 )
    hwerr = 1 ;

return ( hwerr ) ;
}


/*  */
/* Function :           ReadA()
 * Purpose :            The address of the register to read from is set to
 *                      to port_B, the data content is lathced into port_A,
 *                      read from port_A and returned.
 *                      This is the lowest level function.
 * Parameters : (in)    xaddress        address of the register to read from
 * Variables :          adatar          data read from port_A
 * Return :             adatar
 */

int ReadA( int xaddress )
{
int adatar ;         /* data read will be returned here */

/* prepare the bus read address */
outp( pdatareg, xaddress ) ;

/* set the bus read address */
outp( pctrlreg, WRITE_ADDR_B ) ;
outp( pctrlreg, WRITE_ADDR_B | STROBE ) ;
outp( pctrlreg, WRITE_ADDR_B ) ;

/* latch the bus data in the input register */
outp( pctrlreg, LATCH_DATA_A ) ;
outp( pctrlreg, LATCH_DATA_A | STROBE ) ;
outp( pctrlreg, LATCH_DATA_A ) ;

/* select the input register low-nibble for reading */
outp( pctrlreg, READ_DATA_LO ) ;
inp( pstatreg ) ;

/* read and shift it to its place */
adatar = ( inp( pstatreg ) & 0x00F0 ) >> 4 ;

/* select the input register high-nibble for reading */
outp( pctrlreg, READ_DATA_HI ) ;
inp( pstatreg ) ;

/* read and add the two nibbles */
adatar |= ( inp( pstatreg ) & 0x00F0 ) ;

/* return the bus to idle state */
outp( pctrlreg, IDLE ) ;

return ( adatar ) ;
}


/*  */
/* Function :           WriteA()
 * Purpose :            The address of the register to write to is set to
 *                      port_B, data is written into port_A, then port_C is
 *                      used for handshaking to push data on the system bus.
 *                      This is the lowest level function.
 * Parameters : (in)    xaddress        address of the register to write to
 *                      adataw          data to write
 * Variables :          (none)
 * Return :             (function returns nothing)
 */

void WriteA( int xaddress, int adataw )
{

/* prepare the bus write address */
outp( pdatareg, xaddress ) ;

/* set the write bus address */
outp( pctrlreg, WRITE_ADDR_B ) ;
outp( pctrlreg, WRITE_ADDR_B | STROBE ) ;
outp( pctrlreg, WRITE_ADDR_B ) ;

/* prepare the bus write data */
outp( pdatareg, adataw ) ;

/* set the write bus data */
outp( pctrlreg, WRITE_DATA_A ) ;
outp( pctrlreg, WRITE_DATA_A | STROBE ) ;
outp( pctrlreg, WRITE_DATA_A ) ;

/* return bus to idle */
outp( pctrlreg, IDLE ) ;
}


/*  */
/* Handle File Errors */
void ErrorFile( int error )
{
switch( error )
	{
	case 1 :
		sprintf( txtbuf, ">> Can't open input file !" ) ;
		break ;
	case 2 :
		sprintf( txtbuf, ">> Can't open output file !" ) ;
		break ;
	case 3 :
		sprintf( txtbuf, ">> Error processing file !" ) ;
		break ;
	case 4 :
		sprintf( txtbuf, ">> Can't read file !" ) ;
		break ;
	case 5 :
		sprintf( txtbuf, ">> Can't write to file !" ) ;
		break ;
	case 6 :
		sprintf( txtbuf, ">> DISK ERROR : Can't close file !" ) ;
		break ;
	case 7 :
		sprintf( txtbuf, ">> Can't rename file !" ) ;
		break ;
	}

ErrorLine( 1, 1 ) ;
}


/*  */
int SaveTestFlags( int pcbnum, int testnum, int connum, int testflag )
{
int ferr, ix, kx, cx, flag_ok ;
FILE *fhrw ;
char fname[13] = { "TP1D0000.FLG" } ;
char fnameok[13] = { "TP1D0000.OK " } ;
char fnroot[5] = { "TP1D" } ;
char flagfext[5] = { ".FLG" } ;
                    /*           1         2         3         4         5         6 */
                    /* 0123456789012345678901234567890123456789012345678901234567890 */
/* char flagline[44] = { "0000  . . . . . .  . . . . . .  . . . . . ." } ; */
char flagline[62] = { "0000  . . . . . . . . .  . . . . . . . . .  . . . . . . . . ." } ;
/* int tnofs[3] = { 6, 19, 32 }  ; */
int tnofs[3] = { 6, 25, 44 }  ;
char tfc ;

if ( pcbnum < 0 )
	pcbnum = 0 ;
if ( pcbnum > 9999 )
	pcbnum = 9999 ;
fnroot[3] = flag_type ;
sprintf( fname, "%s%04d%s", fnroot, pcbnum, flagfext ) ;
ferr = 0 ;

if ( ( fhrw = fopen( fname, "r+t" ) ) != NULL )
	{
	if ( fgets( txtbuf, MAXSTRING, fhrw ) != NULL )
		{
		txtbuf[44] = '\0' ;
		kx = atoi( txtbuf ) ;
		if ( kx == pcbnum )
			{
			/* find the current flag position */
			cx = tnofs[ testnum - 1 ] + 2 * ( connum - 1 ) ;
			/* if ( ( cx > 5 ) && ( cx < 43 ) ) */
			if ( ( cx > 5 ) && ( cx < 62 ) )
				{
				tfc = txtbuf[cx] ;
				if ( tfc != '1' )
					{
					if ( testflag )
						txtbuf[cx] = '1' ;
					else
						txtbuf[cx] = '0' ;
					sprintf( flagline, "%s", txtbuf ) ;
					fseek( fhrw, 0L, SEEK_SET ) ; /* return to the begining of the file */
					ferr = fputs( flagline, fhrw ) ;
					if ( ferr )
						ErrorFile( 5 ) ;

					/* test all flags ; if all ones, rename the file */
					flag_ok = 1 ;
					for( kx = 1 ; kx < 4 ; kx++ )
						{
						/* for( ix = 1 ; ix < 7 ; ix++ ) */
						for( ix = 1 ; ix < 10 ; ix++ )
							{
							cx = tnofs[ kx - 1 ] + 2 * ( ix - 1 ) ;
							if ( flagline[cx] != '1' )
								{
								flag_ok = 0 ;
								}
							}
						}
					if ( flag_ok )
						{
						fclose( fhrw ) ;
						sprintf( fnameok, "%s%04d.OK", fnroot, pcbnum ) ;
						sprintf( txtbuf, "%s", fnameok ) ; ErrorLine( 0, 1 ) ;
						ferr = rename( fname, fnameok ) ;
						if ( ferr )
							ErrorFile( 7 ) ;
						}
					}
				/* else : error : attempt to overwrite a good test */
				}
			/* else : input data error ! */
			}
		/* else : wrong pcb-number ! */
		}
	else
		{
		/* else : End-Of-File reached, no data read */
		ErrorFile( 4 ) ;
		}
	if ( !flag_ok )
		fclose( fhrw ) ;
	}
else
	{
	if ( ( fhrw = fopen( fname, "w+t" ) ) != NULL )
		{
		sprintf( txtbuf, "%04d", pcbnum ) ;
		for( kx = 0 ; kx < 4 ; kx++ )
			flagline[kx] = txtbuf[kx] ;
		cx = tnofs[ testnum - 1 ] + 2 * ( connum - 1 ) ;
		/* if ( ( cx > 5 ) && ( cx < 43 ) ) */
		if ( ( cx > 5 ) && ( cx < 62 ) )
			{
			if ( testflag )
				flagline[cx] = '1' ;
			else
				flagline[cx] = '0' ;
			ferr = fputs( flagline, fhrw ) ;
			if ( ferr )
				ErrorFile( 5 ) ;
			}
		}
	else
		ErrorFile( 2 ) ;

	if ( fclose( fhrw ) )
		ErrorFile( 6 ) ;
	}

return ferr ;
}


/*  */
int HowManyFiles( void )
{
int kx ;
struct find_t found ;
char fmask[10] = { "TP1D*.TST" } ;

fmask[3] = flag_type ;
kx = 0 ;
if ( !_dos_findfirst( fmask, 0xFFFF, &found ) )
	{
	if ( ( found.size ) > 0L )
		kx++ ;

    while( !_dos_findnext( &found ) )
        {
        if ( ( found.size ) > 0L )
			kx++ ;
		}
    }

return kx ;
}


/*  */
int MakeFileList( int pagelen, int yy0 )
{
int jx, kx, lx ;
struct find_t found;
char fmask[10] = { "TP1D*.TST" } ;

fmask[3] = flag_type ;
kx = HowManyFiles() ;

jx = 0 ;
lx = 0 ;

if( !_dos_findfirst( fmask, 0xFFFF, &found ) )
	{
	if ( ( found.size ) > 0L )
		{
		jx++ ;
		if ( yy0 == 1 )
			{
			sprintf( filelist[lx], "%s", found.name ) ;
			lx++ ;
			}
		}
    while( ( !_dos_findnext( &found ) ) && ( pagelen > lx ) && ( kx > jx ) )
        {
        if ( ( found.size ) > 0L )
			{
			jx++ ;
			if ( jx >= yy0 )
				{
				sprintf( filelist[lx], "%s", found.name ) ;
				lx++ ;
				}
			}
		}
    }

if ( lx < pagelen )
	{
	for( kx = lx ; kx <= pagelen ; kx++ )
		{
		sprintf( filelist[kx], "%s", "" ) ;
		}
	}

return jx ;
}


/*  */
void ScrollFiles( int pagelen, int yy0, int yy1 )
{
int lx ;
char buffer[13] ;

for ( lx = 0 ; lx < pagelen ; lx++ )
	{
	_settextposition( lx+1, 1 ) ;
	sprintf( buffer, "% 12s", filelist[lx] ) ;
	if ( lx == yy1 - yy0 )
		{
		_setbkcolor ( 7L ) ;
		_settextcolor ( 0 ) ;
		}
	else
		{
		_setbkcolor ( 1L ) ;
		_settextcolor ( 7 ) ;
		}
	_outtext ( buffer ) ;
	}

_setbkcolor ( 1L ) ;
_settextcolor ( 7 ) ;
}


/*  */
/* file selection by cursor */

void KeyAction( int key, int ftotal, int pagelen, int *yy0, int *yy1 )
{
/* yy0 points to the top of the filelist[][]      */
/* yy1 points to the current file within the list */

switch ( key )
	{
	case 18432 /* UP */ :
		{
		*yy1 = *yy1 - 1 ;
		if ( *yy1 < ( *yy0 + 1 ) )
			*yy0 = *yy0 - 1 ;
		break ;
		}
	case 20480 /* DOWN */ :
		{
		*yy1 = *yy1 + 1 ;
		/* if ( *yy1 > ( *yy0 + pagelen - 3 ) ) */
		if ( *yy1 > ( *yy0 + pagelen - 2 ) )
			*yy0 = *yy0 + 1 ;
		break ;
		}
	case 18688 /* PGUP */ :
		{
		*yy0 -= pagelen ;
		*yy1 -= pagelen ;
		break ;
		}
	case 20736 /* PGDN */ :
		{
		*yy0 += pagelen ;
		*yy1 += pagelen ;
		break ;
		}
	case 18176 /* HOME */ :
		{
		*yy0 = 1 ;
		*yy1 = 1 ;
		break ;
		}
	case 20224 /* END */ :
		{
		/* *yy0 = ftotal - pagelen + 2 ; */
		*yy0 = ftotal - pagelen + 1 ;
		*yy1 = ftotal ;
		break ;
		}
	default :
		break ;
	}

/* set limits */
if ( *yy0 > ( ftotal - pagelen + 1 ) )
	*yy0 = ftotal - pagelen + 1 ;
if ( *yy0 < 1 )
	*yy0 = 1 ;
if ( *yy1 < 1 )
	*yy1 = 1 ;
if ( *yy1 > ftotal )
	*yy1 = ftotal ;

}

/*  */
int SelectFile( char *filename, int newlist )
{
int x1, y1, x2, y2;
int k;
int a, b, c, d;
int key = 0 ;
int ferr = 0 ;

/* define filelist pointers as static - refresh if newlist flag > 0 */

static int yy0, yy1, yy2 ;

/* yy0 points to the file at the top of the filelist[][] */
/* yy1 points at the highloghted file within the list    */
/* yy2 is the previous value of yy0                      */

int firstfnum, pagelen, ftotal, jx ;
char fmask[10] = { "TP1D*.TST" } ;

fmask[3] = flag_type ;
x1 = 35 ;
y1 = 4 ;
x2 = x1 + 20 ;
y2 = y1 + 18 ;
_settextwindow ( y1, x1, y2, x2 ) ;
_setcolor ( 15 ) ;
_setbkcolor ( 1L ) ;
sprintf ( txtbuf, " files : *.TST ͻ" ) ;
ScrLine( 1, 1 ) ;
for ( k = 2 ; k < ( y2 - y1 ) ; k++ )
	{
	sprintf ( txtbuf, "                   " ) ;
	ScrLine( k, 1 ) ;
	}
sprintf ( txtbuf, "ͼ" ) ;
                /* 123456789012345678901 */
				/*   total: xxxx files   */
ScrLine( k, 1 ) ;

ftotal = HowManyFiles() ;
sprintf( txtbuf, " total: % 4d files ", ftotal );
ScrLine( k, 2 ) ;

a = y1 + 1 ;
b = x1 + 2 ;
c = y2 - 2 ;
d = x2 - 2 ;
pagelen = c - a ;
_settextwindow ( a, b, c, d ) ;
_settextposition ( 1, 1 ) ;
/* _outtext ( "Testing the writing of a long line of text into a text window with the line bending option" ) ; */

if ( ftotal > 0 )
	{
	if ( newlist )
		{
		yy0 = 1 ;
		yy1 = 1 ;
		yy2 = 1 ;
		}
	/* else - use previous values : defined as static ! */

	jx = MakeFileList( pagelen, yy0 ) ;
	ScrollFiles( pagelen, yy0, yy1 ) ;

  	do  {
		yy2 = yy0 ;
		key = WaitKey( 1 ) ;
		KeyAction( key, ftotal, pagelen, &yy0, &yy1 );
		if ( yy0 != yy2 )
			jx = MakeFileList( pagelen, yy0 ) ;
		ScrollFiles( pagelen, yy0, yy1 ) ;
		if ( 0 )
			{
			sprintf(txtbuf, "0=%2d 1=%2d p=%2d", yy0, yy1, pagelen ) ;
			ScrLine( y2-y1, 1 ) ;
			}
		} while ( !( ( key == 283 /* ESC */ ) || ( key == 7181 /* ENTER */ ) ) ) ;

	if ( key == 7181 /* ENTER */ )
		sprintf( filename, "%s", filelist[yy1-yy0] ) ;
	else
		{
		/* no file selected */
		/* sprintf( filename, "%s", '\0' ) ; */
		filename[0] = '\0' ;
		ferr = -2 ;
		}
	}
else
	{
	sprintf( filename, "%s", "" ) ;
	sprintf( txtbuf, "No files !" ) ;
	ScrLine( 2, 2 ) ;
	WaitKey( 1 ) ;
	ferr = -1 ;
	}

/* clear the window text */
_clearscreen( _GWINDOW ) ;

/* return the default colors and clear the file window */
_setbkcolor( 0L ) ;
_setcolor( 7 ) ;
_settextwindow( y1, x1, y2, x2 ) ;
_clearscreen( _GWINDOW ) ;

/* set full-screen window back */
_settextwindow( 1, 1, 25, 80 ) ;

return ferr ;
}


/*  */
void ListFile( char *filename )
{
FILE *fhr ;
int fields, kx, cx, mx, nlines ;
int key ;
char *p_str ;

kx = 0 ;
fields = 0 ;
nlines = 19 ;

if ( ( fhr = fopen( filename, "rt" ) ) != NULL )
	{
	do	{
		if ( fgets( txtbuf, MAXSTRING, fhr ) == NULL )
			{
			fields = EOF ;
			}
		else
			{
			/* find NEWLINE and replace with NULL */
			cx = 0 ;
			while( ( txtbuf[cx] != '\n' ) && cx < MAXSTRING )
				cx++ ;
			txtbuf[cx] = '\0' ;
			ScrLine( 3 + ( ( kx % nlines ) + 1 ), 1 ) ;
			kx++ ;
			fields++ ;
			if ( !( kx % nlines ) )
				{
				/* sprintf( txtbuf, "kx = % 4d     div = % 4d", kx, kx % nlines ) ; */
				/* ScrLine( 1, 40 ) ; */
				_settextposition( 25, 1 ) ;
				_outtext( "Press any key for next page or ESC to exit ..." ) ;
				key = WaitKey( 1 ) ;
				ClearLines( 4, 25 ) ;
				}
			}
		} while( ( fields != EOF ) && ( key != 283 ) ) ;

	fclose( fhr ) ;
	WaitKey( 1 ) ;
	ClearLines( 3, 25 ) ;
	}
else
	ErrorFile( 1 ) ;

}


/*  */
/* Function :		TimeStr()
 * Purpose :		Converts from unsigned time, format FEDCBA9876543210
 *					s=2s incr., m=0-59, h=23            hhhhmmmmmmmsssss
 *					into a 9-byte string (ignore s)     hh:mm ?m  (am/pm)
 * Parameters :     (u)t ....... time
 *					*buf ....... pointer to char.
 * Variables :		h, m
 * Returns :		buf ........ (9-byte string)
 */

char *TimeStr( unsigned t, char *buf )
{
int h, m ;

h = ( t >> 11 ) & 0x1F ;
m = ( t >> 5 ) & 0x3F ;

/* sprintf( buf, "%2.2d:%02.2d %cm", h % 12, m, h > 11 ? 'p' : 'a' ) ; */

/* return only "hh:mm" */
sprintf( buf, "%2.2d:%02.2d", h, m ) ;

return buf ;
}


/*  */
/* Function :		DateStr()
 * Purpose :		Converts from (u)date, format        FEDCBA9876543210
 *					d=1-31, m=1-12, y=1-199 (1980-2099)  yyyyyyymmmmddddd
 *					into a 11-byte string                yyyy/mm/dd
 * Parameters :		(u)d ....... date
 *					*buf ....... pointer to char.
 * Variables :		(none)
 * Return :			buf
 */

char *DateStr( unsigned d, char *buf )
{
sprintf( buf, "%4d-%02.2d-%02.2d", ( d >> 9 ) +1980, ( d >> 5 ) &0x0F, d & 0x1F ) ;

return buf ;
}


/*  */
/* Function :			GetLastFile()
 * Purpose :			Find the last saved file, either by date/time or
 *						by the name-number. Display the file info and
 *						return the name-number.
 * Parameters :			nt_flag ..... number (0) or date/time (1) search
 * Variables :			pp1num ...... file name-number
 *						nextnum ..... next file name-number
 *						kx .......... file counter
 *						ferr ........ file error flag
 *                      *fhp ........ file handle
 *						*finfo ...... pointer to file info (struct find_t)
 *						*fmask ...... file search mask string, "*.TST"
 *						fsize ....... file size
 *						dsize ....... total dir size of matching files
 * Return :				pp1num
 */

int GetLastFile( int nt_flag )
{
int pp1num, nextnum, kx ;
unsigned ftime, xtime, fdate, xdate ;
long fsize, dirsize ;
FILE *fhp ;
struct find_t finfo ;
char timebuf[6], datebuf[12], filename[13], fnroot[5] ;

pp1num = 0 ;
kx = 0  ;
fsize = 0L ;
dirsize = 0L ;
sprintf( filename, "%s", "*.TST" ) ;

if ( !( _dos_findfirst( filename, 0xFFFF, &finfo ) ) )
	{
	kx++ ;
	fsize = finfo.size ;
	dirsize += fsize ;
	pp1num = atoi( &finfo.name[4] ) ;
	fdate = finfo.wr_date ;
	ftime = finfo.wr_time ;
	sprintf( fnroot, "TP1%c", finfo.name[3] ) ;

	while( !( _dos_findnext( &finfo ) ) )
		{
		kx++ ;
		fsize = finfo.size ;
		dirsize += fsize ;
		nextnum = atoi( &finfo.name[4] ) ;
		xdate = finfo.wr_date ;
		xtime = finfo.wr_time ;
		if ( nt_flag )
			{
			if ( xdate > fdate )
				{
				ftime = xtime ;
				fdate = xdate ;
				pp1num = nextnum ;
				sprintf( fnroot, "TP1%c", finfo.name[3] ) ;
				}
			else if ( xdate == fdate )
				{
				if ( xtime > ftime )
					{
					ftime = xtime ;
					fdate = xdate ;
					pp1num = nextnum ;
					sprintf( fnroot, "TP1%c", finfo.name[3] ) ;
					}
				}
			}
		else
			{
			if ( nextnum > pp1num )
				{
				pp1num = nextnum ;
                ftime = xtime ;
                fdate = xdate ;
				sprintf( fnroot, "TP1%c", finfo.name[3] ) ;
				}
			}
		}

	sprintf( filename, "%s%04d.TST", fnroot, pp1num ) ;
	_dos_findfirst( filename, 0xFFFF, &finfo ) ;

	fdate = finfo.wr_date ;
	DateStr( fdate, datebuf ) ;
	ftime = finfo.wr_time ;
	TimeStr( ftime, timebuf ) ;
	fsize = finfo.size ;

	if ( nt_flag )
		sprintf( txtbuf, "Last File (by date/time) :" ) ;
	else
		sprintf( txtbuf, "Last File (by filename)  :" ) ;
	ScrLine( 3, 1 ) ;
	sprintf( txtbuf, "Filename        date        time     size [bytes]" ) ;
	ScrLine( 5, 1 ) ;
	sprintf( txtbuf, "%s%04d.TST   %11s  %5s        % 8ld", fnroot, pp1num, datebuf, timebuf, fsize ) ;
	ScrLine( 6, 1 ) ;
	sprintf( txtbuf, "Total files : % 5d         Total size : % 8ld", kx, dirsize ) ;
	ScrLine( 8, 1 ) ;
	}
else
	{
	sprintf( txtbuf, "WARNING : No *.TST files found !" ) ;
	ErrorLine( 1, 1 ) ;
	}

return pp1num ;
}


/*  */

void dispmenu( int file_flag )
{
int xpos = 57 ;
/* char txtbuf[81]; */

_clearscreen ( _GCLEARSCREEN ) ;
_setbkcolor ( 1L ) ;
_settextcolor ( 15 ) ;

sprintf ( txtbuf, "  Patch-Panel Test Program  " ) ;
ScrLine( 1, 1 ) ;

sprintf ( txtbuf, "  RXon, IJS/F9  " ) ;
ScrLine( 2, 1 ) ;
				/* 5  6         7         8 */
				/* 789012345678901234567890 */
sprintf ( txtbuf, "  Menu ͻ " ) ;
ScrLine( 4, xpos ) ;
sprintf ( txtbuf, "                      " ) ;
ScrLine( 5, xpos ) ;
sprintf ( txtbuf, "  F1 - Hardware Test  " ) ;
ScrLine( 6, xpos ) ;
sprintf ( txtbuf, "  F2 - Capacitors     " ) ;
ScrLine( 7, xpos ) ;
sprintf ( txtbuf, "  F3 - High Voltage   " ) ;
ScrLine( 8, xpos ) ;
sprintf ( txtbuf, "  F4 - Resistance     " ) ;
ScrLine( 9, xpos ) ;
sprintf ( txtbuf, "                      " ) ;
ScrLine( 10, xpos ) ;
sprintf ( txtbuf, "  F5 - File List      " ) ;
ScrLine( 11, xpos ) ;
sprintf ( txtbuf, "  F6 - Last File (N)  " ) ;
ScrLine( 12, xpos ) ;
sprintf ( txtbuf, "  F7 - Last File (T)  " ) ;
ScrLine( 13, xpos ) ;
sprintf ( txtbuf, "  F8 - PCB Type B/F   " ) ;
ScrLine( 14, xpos ) ;
sprintf ( txtbuf, "                      " ) ;
ScrLine( 15, xpos ) ;
sprintf ( txtbuf, "  F9 -                " ) ;
ScrLine( 16, xpos ) ;
sprintf ( txtbuf, "  F10-                " ) ;
ScrLine( 17, xpos ) ;
sprintf ( txtbuf, "                      " ) ;
ScrLine( 18, xpos ) ;

if ( !file_flag )
	{
	sprintf ( txtbuf, "                      " ) ;
	ScrLine( 19, xpos ) ;
	sprintf ( txtbuf, "                      " ) ;
	ScrLine( 20, xpos ) ;
	sprintf ( txtbuf, "                      " ) ;
	ScrLine( 21, xpos ) ;
	}
else
	{
	sprintf ( txtbuf, "  %c %c PgUp PgDn       ", 0x18, 0x19 ) ;
	ScrLine( 19, xpos ) ;
	sprintf ( txtbuf, "  Home  End   ESC     " ) ;
	ScrLine( 20, xpos ) ;
	sprintf ( txtbuf, "  Enter = Select      " ) ;
	ScrLine( 21, xpos ) ;
	}
sprintf ( txtbuf, "                      " ) ;
ScrLine( 22, xpos ) ;
sprintf ( txtbuf, "  ESC - Exit ͼ " ) ;
ScrLine( 23, xpos ) ;

/* reset colors to default */
_setbkcolor ( 0L ) ;
_settextcolor ( 7 ) ;
}


/*  */
/* Function :           ReadDTStat()
 * Purpose :            Read the Status Register of the Delta_t card
 *						Bit :           Description :
 *                        0                Mode_0
 *                        1                Mode_1
 *                        2                Polarity-A
 *                        3                Polarity-B
 *                        4                ClkEnable=1
 *                        5                WaitTrig=1
 *                        6                DataReady=1
 *                        7                OverRange=1
 * Parameters :         (none)
 * Variables :          stat
 * Return :             stat
 */

int ReadDTStat( void )
{
int stat ;

stat = ReadA( 0x07 ) ;

return( stat ) ;
}


/*  */
/* Function :           SetDTMode()
 * Purpose :            Set Delta_t card Mode-Register
 *						Bit :           Description :
 *                        0                Mode_0
 *                        1                Mode_1
 *                        2                Polarity-A
 *                        3                Polarity-B
 *                        4                Trig-A invert=1
 *                        5                Trig-B invert=1
 *                        6                not used
 *                        7                not used
 * Parameters :		ab_mode
 * Mode Parameter :
 ***************************************
 *					0x00 = A+ to B+
 *					0x01 = B+ to A+
 *					0x20 = A- to B+
 *					0x21 = B+ to A-
 *					0x10 = A+ to B-
 *					0x11 = B- to A+
 *					0x30 = A- to B-
 *					0x31 = B- to A-
 ***************************************
 *					0x02 = A+ to next A+
 *					0x0A = A+ to A-
 *					0x22 = A- to next A-
 *					0x2A = A- to A+
 ***************************************
 *					0x03 = B+ to next B+
 *					0x07 = B+ to B-
 *					0x13 = B- to next B-
 *					0x17 = B- to B+
 ***************************************
 * Variables :          (none)
 * Return :             (function returns nothing)
 */

void SetDTMode( int ab_mode )
{

WriteA( 0x06, ab_mode ) ;

}


/*  */
/* Function :           StartNewDTM()
 * Purpose :            Start new Delta_t Measurement
 * Parameters :         (none)
 * Variables :          (none)
 * Return :             (function returns nothing)
 */

void StartNewDTM( void )
{

/* start new measurement */
WriteA( 0x04, 0x00 ) ;

}


/*  */
/* Function :           ResetDT()
 * Purpose :            Reset Delta_t System
 * Parameters :         (none)
 * Variables :          (none)
 * Return :             (function returns nothing)
 */

void ResetDT( void )
{

/* reset Delta_t card */
WriteA( 0x05, 0x00 ) ;

}


/*  */
/* Function :           ReadDTCounter()
 * Purpose :            Read the 24-bit counter, resolution 100ns
 * Parameters : (in)    *p_cnt (pointer to unsigned long)
 * Variables :          data_high	(int)
 *						data_mid	(int)
 *						data_low	(int)
 *						delta_t		(unsigned long)
 * Return :             (function returns nothing)
 */

void ReadDTCounter( unsigned long *p_cnt )
{
unsigned int data_low, data_mid, data_high ;
unsigned long int delta_t ;

/* latch-in data and read high-byte */
data_high = ReadA( 0x06 ) ;

/* read mid-byte */
data_mid = ReadA( 0x05 ) ;

/* read low-byte */
data_low = ReadA( 0x04 ) ;


delta_t = data_high ;
delta_t <<= 16 ;
delta_t |= ( data_mid << 8 ) ;
delta_t |= data_low ;

*p_cnt = delta_t ;
}


/*  */
/* Function :			ResetCMS()
 */

void ResetCMS( void )
{
/* disable HC238, D4=1, set bits 6 and 7, clear other bits */
WriteA( 0x00, 0xD0 ) ;

/* reset the DataReadyFlag flip-flop */
WriteA( 0x01, 0x00 ) ;

/* wait 55ms, ignore flag_kb */
Delay55ms( 0 ) ;

/* disable HC238, D4=1, clear all other bits */
WriteA( 0x00, 0x10 ) ;
}


/*  */
/* Function :			ReadCMS()
 */

int ReadCMS( void )
{
int comp16 ;

comp16 = ReadA( 0x03 ) ;
comp16 <<= 8 ;
comp16 |= ReadA( 0x02 ) ;
comp16 = ~comp16 ;

return comp16 ;
}


/*  */
/* Function :			ReadCMStat()
 */

int ReadCMStat( void )
{
int cmstat ;

cmstat = ReadA( 0x00 ) ;

return cmstat ;
}


/*  */
/* Function :			StartCMS()
 */

void StartCMS( int channel )
{
/* Write to Latch Address 0x00 ;
 * HC74  start bit D5=1, [Data | 0x20] ( cleared by ResetCMS() ),
 * HC238 enable bit D4=0, [Data & 0xF7] ( set by ResetCMS() ),
 * Low-8/High-8 select bit D3, 0=L, 1=H ;
 * 1-of-8 selection with bits D2, D1, D0 ;
 * Note : bit D7=High shorts ch.15 to GND ;
 *        bit D6=High shorts ch.14 to GND ;
 */

WriteA( 0x00, ( 0x20 | channel ) ) ;

}


/*  */
/* Function :			ReadADC1601()
 * Purpose :			Set one of 8 input channels, start the conversion
 *						and read the AD7805 16-bit AD converter data
 * Parameters :		(in) chnum ....... the channel number, 0-7
 * Variables :			adcdata ...... the 16-bit ADC data
 *						hwerr ........ hardware response error flag
 *						loop ......... loop count for error detection
 * Return :				adcdata
 */

int ReadADC1601( int chnum, int flag_dly )
{
int adcdata, status, loop ;
/* char txtbuf[80] ; */

loop = 0 ;
adcdata = 0 ;

/* set ADC channel */
WriteA( 0x08, chnum ) ;
if ( flag_dly )
	Delay55ms( 0 ) ;

/* start convert */
WriteA( 0x09, 0 ) ;

/* check _BUSY ( max. 10x ) */
status = 0x40 & ReadA( 0x08 ) ;
while( ( !status ) && !( loop / 10 ) )
	{
	status = 0x40 & ReadA( 0x08 ) ;
	loop++ ;
	}

if ( loop && !status )
	{
    sprintf( txtbuf, ">> Error ADC1601 : loop = 0x%02X - no_BUSY !", loop ) ;
	ErrorLine( 1, 1 ) ;
	}
else
	{
	/* read ADC */
	adcdata = ReadA( 0x0B ) ;
	adcdata |= ( ReadA( 0x0A ) << 8 ) ;
	}

return adcdata ;
}


/*  */
/* Function :			HVTestInit()
 * Purpose :			High-Voltage Test card initialization and
 *						power supply diagnostics.
 * Parameters :			calibr ...... calibration request flag
 *						*Vground .... output ground level
 *						*Gainadc .... output ADC gain
 * Variables :			hwerr ....... hardware response error flag
 *						testcode .... latch presence R/W test code
 *						chnum ....... ADC channel number, 0-7
 *						hvonoff ..... High-Voltage ON/OFF control bit
 *						status ...... Mode latch and ADC _BUSY status
 *						adcdata ..... the 16-bit ADC data
 *						Vadc ........ adcdata converted to voltage
 *						Vgnd ........ ground voltage for offset correcion
 *						Gnorm ....... ADC gain ( 2.5/Vref )
 *
 * Return :				hwerr
 */

int HVTestInit( int calibr, double *Vground, double *Gainadc )
{
unsigned testcode ;
int hwerr, flag_kb ;
int kx, chnum, hvonoff, status ;
int adcdata, Navg ;
/* char txtbuf[80] ; */

double Vadc, Vgnd, Gnorm ;

hwerr = 0 ;
flag_kb = 0 ;
testcode = 0x24 ;
chnum = 0 ;
hvonoff = 0 ;
Navg = 256 ;

Vgnd = 0.0 ;
Gnorm = 1.0 ;

/* Write address 0x08 :
 *		Latch 74HC574 Control Register
 *		MSB bit D7 = 1 ==> High Voltage ON ;
 *       bits D2,D1,D0 ==>  ADC channel select :
 *       (0)  0  0  0  ==>  Gnd ref.level ;
 *       (1)  0  0  1  ==>  ADC Vref=10Vdc, code : 0x7FFF ;
 *       (2)  0  1  0  ==>  +12V * ( 13 / ( 13 + 18 ) ) ;
 *       (3)  0  1  1  ==>  -12V * ( 13 / ( 13 + 18 ) ) ;
 *       (4)  1  0  0  ==>  Gnd : I_leak * 10Mohm ;
 *       (5)  1  0  1  ==>  HVRet : I_leak * 1Mohm * 10 ;
 *       (6)  1  1  0  ==>  HVBias / 101 ;
 *       (7)  1  1  1  ==>  +5Vdc ;
 *
 * Write address 0x09 :
 *		ADC ADS7805 start convert pulse : Low min 40ns, max 8us ;
 *
 * Read address 0x08 :
 *		Latch and read Status : D0-D5 and D7 input from Control Reg.;
 *								D6 : ADC _BUSY ;
 *
 * Read address 0x0B :
 *		Latch ADC 16-bit data and read LSByte ;
 *
 * Read address 0x0A :
 *		Read ADC MSByte ;
 */

if ( !( calibr & 1 ) )
	{
	/* do this for calibr == 0 or 2 */

	/* write the testcode */
	WriteA( 0x08, ( chnum | hvonoff | testcode ) ) ;

	/* test the latch content, but ignore D6 : ADC _BUSY */
	hwerr = 0xBF & ReadA( 0x08 ) ;
	if ( hwerr == testcode )
		hwerr = 0 ;

	/* printf( "\n>> 1 : hwerr = 0x%02X", hwerr ) ; */
	WriteA( 0x08, ( chnum | hvonoff ) ) ;
	}

if ( ( calibr & 1 ) && !hwerr )
	{
	/* do this for calibr == 1 */

	/* read ADC channel 0 : GND = 0 */
	chnum = 0 ;
	Vadc = 0.0 ;
	adcdata = ReadADC1601( chnum, 1 ) ;
	/* Delay55ms( 0 ) ; */
	for ( kx = Navg ; kx > 0 ; kx-- )
		{
		adcdata = ReadADC1601( chnum, 0 ) ;
        Vadc += ( (double) adcdata ) / 3276.8 ;
		}
	/* remember GND value for correction */
	Vgnd = Vadc / (double)Navg ;
	*Vground = Vgnd ;
	/* printf( "\n>> Vgnd = %+ 7.3fV", Vgnd ) ; */

	/* read ADC channel 1 : Vref = +2.5 V */
	chnum = 1 ;
	Vadc = 0.0 ;
	adcdata = ReadADC1601( chnum, 1 ) ;
	/* Delay55ms( 0 ) ; */
	for ( kx = Navg ; kx > 0 ; kx-- )
		{
		adcdata = ReadADC1601( chnum, 0 ) ;
		Vadc += ( (double) adcdata ) / 3276.8 ;
		}
	Vadc = ( Vadc / (double)Navg ) - Vgnd ;
	/* printf( "\n>> Vref = %+ 7.3fV", Vadc ) ; */
	/* normalize gain error */
	Gnorm = 2.5 / Vadc ;
	*Gainadc = Gnorm ;

	if ( fabs( Vgnd ) > 0.010 )
		hwerr |= 0x1000 ;
	if ( fabs( Gnorm - 1.0 ) > 0.050 )
		hwerr |= 0x2000 ;
	}

if ( !hwerr && !calibr )
	{
	/* do this for calibr == 0 */
	/* retrieve calibration data from the calling environment (main()?) */
	Gnorm = *Gainadc ;
	Vgnd = *Vground ;

	/* set ADC channel to +5V supply */
	chnum = 7 ;
	adcdata = ReadADC1601( chnum, 1 ) ;
	Vadc = ( (double) adcdata ) / 3276.8 ;
	Vadc = Gnorm * ( Vadc - Vgnd ) ;
	sprintf( txtbuf, ">> HV-Test card  +5Vdc :  Vadc = %+ 7.3fV", Vadc ) ;
	ScrLine( 17, 1 ) ;

	/* set ADC channel to +12V supply */
	chnum = 2 ;
	adcdata = ReadADC1601( chnum, 1 ) ;
	Vadc = ( (double) adcdata ) / 3276.8 ;
	Vadc = Gnorm * ( Vadc - Vgnd ) ;
	/* correct for the supply divider resistors */
	Vadc = Vadc * ( 13.0 + 18.0 ) / 13.0 ;
	sprintf( txtbuf, ">> HV-Test card +12Vdc :  Vadc = %+ 7.3fV", Vadc ) ;
	ScrLine( 18, 1 ) ;

	/* set ADC channel to -12V supply */
	chnum = 3 ;
	adcdata = ReadADC1601( chnum, 1 ) ;
	Vadc = ( (double) adcdata ) / 3276.8 ;
	Vadc = Gnorm * ( Vadc - Vgnd ) ;
	/* correct for the supply divider resistors */
	Vadc = Vadc * ( 13.0 + 18.0 ) / 13.0 ;
	sprintf( txtbuf, ">> HV-Test card -12Vdc :  Vadc = %+ 7.3fV", Vadc ) ;
	ScrLine( 19, 1 ) ;
	}

return hwerr ;
}


/*  */
/* Function :			HVTestRun()
 *
 */

int HVTestRun( int pp1num, int connum, int timx )
{
int key, flag_kb, ferr, hwerr, testerr, testflag ;
int hvonoff, adcdata, adcchnum, Navg ;
int firsttime, timeout_flag, kx ;
float tzero, tx ;
double Vground, Gainadc, Vadc, HVBias, Ilkret, Ilkgnd, ilkgmax, ilkgmin ;
char filename[13] = { "TP1D0000.TST" } ;
char filenext[5] = { ".TST" } ;
char fnroot[5] = { "TP1D" } ;
FILE *fhp ;
char d_str[9], t_str[9] ;

hwerr = 0 ;
flag_kb = 0 ;
firsttime = 1 ;
Navg = 256 ;
fhp = '\0' ;
testerr = 0 ;
testflag = 1 ;
fnroot[3] = flag_type ;
ilkgmax = 100.0 ; /* ***** max. leakage current = 100 nA ***** */
ilkgmin = 0.2 * ilkgmax ; /* min. leakage during first 10s after HVon */

/* calibrate ADC by first sampling Vgnd, then Vref */
hwerr = HVTestInit( 1, &Vground, &Gainadc ) ;

if ( timx )
	{
	/* write pp1num to the last four filename characters */
	sprintf( filename, "%s%04d%s", fnroot, pp1num, filenext ) ;

	/* open the file to which the test results are going to be written */
	fhp = fopen( filename, "a+t" ) ;
	if ( fhp == '\0' )
		{
		ErrorFile( 1 ) ;
		}
	}

if ( !hwerr )
	{
	/* switch HV ON */
	hvonoff = 0x80 ;
	WriteA( 0x08, hvonoff ) ;
	Delay55ms( 6 ) ;

	if ( !timx )
		{
        /* set ADC channel to PCBgnd */
		adcchnum = 4 ;
		Ilkgnd = 0.0 ;
		adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
		/* Delay55ms( 0 ) ; */
		for ( kx = Navg ; kx > 0 ; kx-- )
			{
			adcdata = ReadADC1601( adcchnum | hvonoff, 0 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkgnd += -Vadc * 1000.0 ;
			}
		Ilkgnd /= (double)Navg ;
		sprintf( txtbuf, ">> HV-Test card : Ilkgnd = %+ 7.1fnA", Ilkgnd ) ;
		ScrLine( 22, 1 ) ;

		/* set ADC channel to HVret */
		adcchnum = 5 ;
		Ilkret = 0.0 ;
		adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
		/* Delay55ms( 0 ) ; */
		for ( kx = Navg ; kx > 0 ; kx-- )
			{
			adcdata = ReadADC1601( adcchnum | hvonoff, 0 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkret = -Vadc * 1000.0 ;
			}
		Ilkret /= (double)Navg ;
		sprintf( txtbuf, ">> HV-Test card : Ilkret = %+ 7.1fnA", Ilkret ) ;
		ScrLine( 23, 1 ) ;

		/* set ADC channel to HVBias */
		adcchnum = 6 ;
		HVBias = 0.0 ;
		adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
		/* Delay55ms( 0 ) ; */
		for ( kx = Navg ; kx > 0 ; kx-- )
			{
			adcdata = ReadADC1601( adcchnum | hvonoff, 0 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 50M/333k divider resistors */
			HVBias += Vadc * (1.0 + 50.0 / 0.333 ) * ( 459.0 / 447.3 ) ;
			}
		HVBias /= (double)Navg ;
		sprintf( txtbuf, ">> HV-Test card : HVBias = %+ 7.1fV", HVBias ) ;
		ScrLine( 21, 1 ) ;

		hvonoff = 0x00 ;
		WriteA( 0x08, hvonoff ) ;
		}
	else
		{
		sprintf( txtbuf, "HV-Test : board = %04d : socket = %1d ", pp1num, connum ) ;
		ScrLine( 4, 1 ) ;

		if ( fhp != '\0' )
			{
			ferr = fprintf( fhp, "\n%s\n", txtbuf ) ;
			ferr = fflush( fhp ) ;
			if ( ferr == EOF )
				{
				ErrorFile( 2 ) ;
				}
			_strdate( d_str ) ;
            _strtime( t_str ) ;
			sprintf( txtbuf, "-------------------- %s - %s -------", d_str, t_str ) ;
			ferr = fprintf( fhp, "%s\n", txtbuf ) ;
			ferr = fflush( fhp ) ;
			if ( ferr == EOF )
				{
				ErrorFile( 2 ) ;
				}
			}

		sprintf( txtbuf, ">> 500V 30s Sustain and Leakage Test : " ) ;
		ScrLine( 7, 1 ) ;

		if ( fhp != '\0' )
			{
			ferr = fprintf( fhp, "%s\n", txtbuf ) ;
			ferr = fflush( fhp ) ;
			if ( ferr == EOF )
				{
				ErrorFile( 2 ) ;
				}
			}

		ClearKbBuf() ;
		flag_kb = 0 ;
		tzero = 0.0 ;
		tx = 25.0 ;
		timeout_flag = CheckTime( &tx, &tzero ) ;
		while( ( !flag_kb ) && ( !timeout_flag ) )
			{
			/* set ADC channel to PCBgnd */
			adcchnum = 4 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkgnd = -Vadc * 1000.0 ;

			/* set ADC channel to HVret */
			adcchnum = 5 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkret = -Vadc * 1000.0 ;

			/* set ADC channel to HVBias */
			adcchnum = 6 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 50M/333k divider resistors */
			Vadc = Vadc * (1.0 + 50.0 / 0.333 ) ;
			/* actuall value correction */
			HVBias = Vadc * 459.0 / 447.3 ;

			sprintf( txtbuf, " HVBias = %+ 7.1f V  ", HVBias ) ;
			ScrLine( 9, 4 ) ;
			sprintf( txtbuf, " IHVret = %+ 7.1f nA ", Ilkret ) ;
			ScrLine( 10, 4 ) ;
			sprintf( txtbuf, "   Ignd = %+ 7.1f nA ", Ilkgnd ) ;
			ScrLine( 11, 4 ) ;

			timeout_flag = CheckTime( &tx, &tzero ) ;
			flag_kb = kbhit() ;
			if ( firsttime && ( tx < 20.0 ) )
				{
				if ( ( fabs( Ilkret ) < ilkgmin ) || ( fabs( Ilkgnd ) < ilkgmin ) )
					{
					testerr = 1 ;
					sprintf( txtbuf, "WARNING : Low cross-charge, < % 3.0f nA : Disconnected ???", ilkgmin ) ;
					ErrorLine( 1, 0 ) ;
					if ( fhp != '\0' )
						{
						ferr = fprintf( fhp, "%s\n", txtbuf ) ;
						ferr = fflush( fhp ) ;
						if ( ferr == EOF )
							{
							ErrorFile( 2 ) ;
							}
						}
					}
				if ( HVBias < 450.0 )
					{
					testerr = 1 ;
					sprintf( txtbuf, "WARNING : Low output voltage : % +7.1f V !", HVBias ) ;
					ErrorLine( 1, 0 ) ;
					if ( fhp != '\0' )
						{
						ferr = fprintf( fhp, "%s\n", txtbuf ) ;
						ferr = fflush( fhp ) ;
						if ( ferr == EOF )
							{
							ErrorFile( 2 ) ;
							}
						}
					}
				firsttime = 0 ;
				}
			}
		if ( flag_kb )
			{
			flag_kb = 0 ;
			ClearKbBuf() ;
			}

		sprintf( txtbuf, "  HVBias [V]    IHVret [nA]    Ignd [nA]" ) ;
		if ( fhp != '\0' )
			{
			ferr = fprintf( fhp, "%s\n", txtbuf ) ;
			ferr = fflush( fhp ) ;
			if ( ferr == EOF )
				{
				ErrorFile( 2 ) ;
				}
			}

		sprintf( txtbuf, "   %+ 7.1f       %+ 7.1f      %+ 7.1f", HVBias, Ilkret, Ilkgnd ) ;
		if ( fhp != '\0' )
			{
			ferr = fprintf( fhp, "%s\n", txtbuf ) ;
			ferr = fflush( fhp ) ;
			if ( ferr == EOF )
				{
				ErrorFile( 2 ) ;
				}
			}
		if ( ( fabs( Ilkret ) > ilkgmax ) || ( fabs( Ilkgnd ) > ilkgmax ) )
			{
			testerr = 1 ;
			sprintf( txtbuf, "WARNING : High leakage, > % 3.0f nA !", ilkgmax ) ;
			ErrorLine( 1, 0 ) ;
			}
		if ( HVBias < 450.0 )
			{
			testerr = 1 ;
			sprintf( txtbuf, "WARNING : Low output voltage : % +7.1f V !", HVBias ) ;
			ErrorLine( 1, 0 ) ;
			}

		timeout_flag = -1 ;
		sprintf( txtbuf, ">> Average values for 5s ... " ) ;
		ScrLine( 13, 1 ) ;

		tzero = 0.0 ;
		tx = 5.0 ;
		timeout_flag = CheckTime( &tx, &tzero ) ;
		kx = 0 ;
		HVBias = 0.0 ;
		Ilkgnd = 0.0 ;
		Ilkret = 0.0 ;
		while( ( !flag_kb ) && ( !timeout_flag ) )
			{
			/* set ADC channel to PCBgnd */
			adcchnum = 4 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkgnd += -Vadc * 1000.0 ;

			/* set ADC channel to HVret */
			adcchnum = 5 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkret += -Vadc * 1000.0 ;

			/* set ADC channel to HVBias */
			adcchnum = 6 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 50M/333k divider resistors */
			Vadc = Vadc * ( 1.0 + 50.0 / 0.333 ) ;
			/* actuall value correction */
			HVBias += Vadc * 459.0 / 447.3 ;

			kx++ ;

			sprintf( txtbuf, " HVBias = %+ 7.1f V  ", HVBias / (double)kx ) ;
			ScrLine( 15, 4 ) ;
			sprintf( txtbuf, " IHVret = %+ 7.1f nA ", Ilkret / (double)kx ) ;
			ScrLine( 16, 4 ) ;
			sprintf( txtbuf, "   Ignd = %+ 7.1f nA ", Ilkgnd / (double)kx ) ;
			ScrLine( 17, 4 ) ;

			timeout_flag = CheckTime( &tx, &tzero ) ;
			flag_kb = kbhit() ;
			}
		if ( flag_kb )
			{
			flag_kb = 0 ;
			ClearKbBuf() ;
			}

		HVBias /= (double)kx ;
		Ilkgnd /= (double)kx ;
		Ilkret /= (double)kx ;

		sprintf( txtbuf, "avg HVBias = %+ 7.1f V  ", HVBias ) ;
		ScrLine( 15, 1 ) ;
		sprintf( txtbuf, "avg IHVret = %+ 7.1f nA ", fabs( Ilkret ) ) ;
		ScrLine( 16, 1 ) ;
		sprintf( txtbuf, "avg   Ignd = %+ 7.1f nA ", fabs( Ilkgnd ) ) ;
		ScrLine( 17, 1 ) ;

		sprintf( txtbuf, "   %+ 7.1f       %+ 7.1f      %+ 7.1f ", HVBias, Ilkret, Ilkgnd ) ;
		if ( fhp != '\0' )
			{
			ferr = fprintf( fhp, ">> 5s average :\n%s\n", txtbuf ) ;
			ferr = fflush( fhp ) ;
			if ( ferr == EOF )
				{
				ErrorFile( 2 ) ;
				}
			}

		if ( ( fabs( Ilkret ) > ilkgmax ) || ( fabs( Ilkgnd ) > ilkgmax ) )
			{
			testerr = 1 ;
			sprintf( txtbuf, "WARNING : High leakage, > % 3.0f nA !", ilkgmax ) ;
			ErrorLine( 1, 0 ) ;
			}
		if ( HVBias < 450.0 )
			{
			testerr = 1 ;
			sprintf( txtbuf, "WARNING : Low output voltage : % +7.1f V !", HVBias ) ;
			ErrorLine( 1, 0 ) ;
			}

		/* clear time display */
		_settextposition( 6, 35 ) ;
		_outtext( "                     " ) ;

		sprintf( txtbuf, ">> HV OFF : Wait untill HVBias falls below 10V ... " ) ;
		ScrLine( 19, 1 ) ;
		/* switch HV OFF */
		hvonoff = 0x00 ;
		WriteA( 0x08, hvonoff ) ;
		ClearKbBuf() ;
		timeout_flag = 0 ;

		while( ( !flag_kb ) && ( !timeout_flag ) )
			{
			/* set ADC channel to PCBgnd */
			adcchnum = 4 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkgnd = -Vadc * 1000.0 ;

			/* set ADC channel to HVret */
			adcchnum = 5 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 1Mohm current sampling resistor and inverting amp. */
			Ilkret = -Vadc * 1000.0 ;

			/* set ADC channel to HVBias */
			adcchnum = 6 ;
			adcdata = ReadADC1601( adcchnum | hvonoff, 1 ) ;
			Vadc = ( (double) adcdata ) / 3276.8 ;
			Vadc = Gainadc * ( Vadc - Vground ) ;
			/* correct for the 50M/333k divider resistors */
			Vadc = Vadc * (1.0 + 50.0 / 0.333 ) ;
			/* actuall value correction */
			HVBias = Vadc * 459.0 / 447.3 ;

			sprintf( txtbuf, " HVBias = %+ 7.1f V  ", HVBias ) ;
			ScrLine( 21, 4 ) ;
			sprintf( txtbuf, " IHVret = %+ 7.1f nA ", Ilkret ) ;
			ScrLine( 22, 4 ) ;
			sprintf( txtbuf, "   Ignd = %+ 7.1f nA ", Ilkgnd ) ;
			ScrLine( 23, 4 ) ;

			/* Don't check the time, check the HV instead ! */
			if ( HVBias > 10.0 )
				timeout_flag = 0 ;
			else
				{
				timeout_flag = 1 ;
				printf( "\a" ) ;
				}
			flag_kb = kbhit() ;
			}
		if ( flag_kb )
			{
			flag_kb = 0 ;
			ClearKbBuf() ;
			}

		/* ***** save test flags ***** */
		if ( !ferr )
			{
			if ( flag_kb || testerr )
				testflag = 0 ;
			ferr = SaveTestFlags( pp1num, 2, connum, testflag ) ;
			}
		}
	}

if ( fhp != '\0' )
	{
	if ( fclose( fhp ) == EOF )
		{
		ErrorFile( 6 ) ;
		}
	}

/* force HV OFF again, just in case */
hvonoff = 0x00 ;
WriteA( 0x08, hvonoff ) ;
if ( flag_kb ) ;
	{
	hwerr = 0 ;
	ClearKbBuf() ;
	}

return hwerr ;
}


/*  */
/* Function :			ReadADC1602()
 *
 */

int ReadADC1602( int *dataout )
{
int hwerr, kx, flag_drdy, adcdata ;
/* char txtbuf[80] ; */

hwerr = 0 ;
kx = 20 ;
flag_drdy = 0x80 ;

/* start convert */
WriteA( 0x0D, 0x00 ) ;

/* test if _Busy goes low and wait for DataReady */
flag_drdy = 0x80 & ReadA( 0x0C ) ;
if ( !flag_drdy )
	{
	kx = 20 ;
	while( kx && !flag_drdy )
		{
		flag_drdy = 0x80 & ReadA( 0x0C ) ;
		kx-- ;
		}
	}

if ( ( !flag_drdy ) && ( !kx ) )
	{
	hwerr |= 0x20 ;
	sprintf( txtbuf, ">> ERROR : ADC reading problems !   hwerr = 0x%02X", hwerr ) ;
	ErrorLine( 1, 1 ) ;
	}

adcdata = ReadA( 0x0F ) ;
adcdata <<= 8 ;
adcdata |= ReadA( 0x0E ) ;
*dataout = adcdata ;

return hwerr ;
}


/*  */
/* Function :			MiliOhmInit()
 * Purpose :			Initialize and test the mOhm meter board
 * Sub Calls :			ReadADC1602()
 * Parameters :			(in) qmode : 10 - test the 10 Ohm internal ref.
 *									  4 - return calibration data, no disp.
 *									  2 - test the status reg. R/W
 *									  1 - test ADC start-conversion
 *									  0 - test ADC calibration
 *						(out) Vground2 .... ADC ground offset level
 *						(out) Vgain2 ...... ADC gain normalized by Vref
 * Variables :			hwerr ........ hardware error
 *						kx, nx ....... counting vars
 *						Navg ......... number of samples averaged
 *						flag_kb ...... keyboard flag
 *						flag_drdy .... ADC data ready flag
 *						adcdata ...... 16-bit int ADC data
 *						Vxxxx ........ voltage values, double-prec.f.p.
 *
 * Return :				hwerr : 0x80 - card not present
 *								0x40 - card not present
 *								0x20 - no _Busy signal
 */

int MiliOhmInit( int qmode, double *Vground2, double *Vgain2 )
{
int hwerr, kx, nx, Navg, flag_kb, flag_drdy, anydata ;
int adcdata ;
double Vgnd, Vadc, Vgain ;

hwerr = 0 ;
kx = 20 ;
nx = 256 ;
Navg = 256 ;
flag_drdy = 1 ;
Vgain = 1.0 ;
Vgnd = 0.0 ;

/* _CS address function description :
 *
 *		Range : 0x0C - 0x0F
 *	************************************************
 *	Write 0x0C : channel selection latch
 *				 D7 ... not used
 *				 D6 ... not used
 *				 D5 ... 1 = enable channels 17-19 + test ch.(20-24)
 *							ch. 20 ( data 0x23 ) ... 10 Ohm reference test
 *							ch. 21 ( data 0x24 ) ... GND reference level, offset (*5)
 *							ch. 22 ( data 0x25 ) ... ADC Vref 2.5V (/2,*5) = +6.25V
 *							ch. 23 ( data 0x26 ) ... +12V (/10,*5) = +6.0V
 *							ch. 24 ( data 0x27 ) ... -12V (/10,*5) = -6.0V
 *				 D4 ... 1 = enable channels  9-16
 *				 D3 ... 1 = enable channels  1-8
 *		   D2,D1,D0 ... select 1 of 8 enabled channels
 *
 *		   Ref. Current :	Iq = 0.1 A
 *		   Res.Equation :	 R = ( 0.2 * Vx ) / Iq
 *
 *	************************************************
 *	Write 0x0D : ADC ADS7805 Start-convert pulse (low min 40ns, max 6000ns)
 *				 Data can be ignored.
 *	************************************************
 *	Write 0x0E : not used
 *	Write 0x0F : not used
 *	************************************************
 *	Read  0x0C : read status register :
 *				 D0-D5 == D0-D5 of channel selection latch
 *				 D6 ... not used, always 0
 *				 D7 ... 0 = conversion started, 1 = data ready
 *	************************************************
 *	Read  0x0D : not used
 *	************************************************
 *	Read  0x0E : ADC ADS7805 LSB data read (latched by rising _Busy)
 *	Read  0x0F : ADC ADS7805 MSB data read (latched by rising _Busy)
 */

/* select channel 21, GND */
WriteA( 0x0C, 0x24 ) ;
Delay55ms( 0 ) ;

if ( qmode == 2 )
	{
	/* check card presence by r/w the status register */
	WriteA( 0x0C, 0x25 ) ;
	anydata = 0x3F & ReadA( 0x0C ) ;
	if ( anydata != 0x25 )
		hwerr |= 0x80 ;
	WriteA( 0x0C, 0x24 ) ;
	anydata = 0x3F & ReadA( 0x0C ) ;
	if ( anydata != 0x24 )
		hwerr |= 0x40 ;
	}

if ( ( qmode & 1 ) && !hwerr )
	{
	/* Check ADC conversion */

	/* select channel 21, GND */
	WriteA( 0x0C, 0x24 ) ;
	/* wait 55ms for the input voltage to stabilize */
	Delay55ms( 0 ) ;

    kx = 20 ;
	/* start conversion */
	WriteA( 0x0D, 0x00 ) ;

	/* test if _Busy goes low and wait for DataReady */
	flag_drdy = 0x80 & ReadA( 0x0C ) ;
	if ( !flag_drdy )
		{
		kx = 20 ;
		while( kx && !flag_drdy )
			{
			flag_drdy = 0x80 & ReadA( 0x0C ) ;
			kx-- ;
			}
		}
	if ( ( !flag_drdy ) && ( !kx ) )
		{
		hwerr |= 0x20 ;
		/* printf( "\n\n>> ERROR : _BUSY low - cannot start ADC !" ) ; */
		/* printf( "\n\n>> flag_dr = 0x%02X\n>> kx = %d\n\n", flag_dr, kx ) ; */
		}
	}

if ( !hwerr )
	{
	/* do this for every qmode */
	/* for qmode = 4, no output to screen */

	/* select channel 21, GND */
	WriteA( 0x0C, 0x24 ) ;
	/* wait 55ms for the input voltage to stabilize */
	Delay55ms( 0 ) ;

    /* ADC calibration, average Vgnd */
	Vgnd = 0.0 ;
	for ( nx = Navg ; nx > 0 ; nx-- )
		{
		hwerr = ReadADC1602( &adcdata ) ;

		/* exit the "for" loop at error */
		if ( hwerr )
			nx = 0 ;

		Vgnd += 0.2 * ( (double)adcdata / 3276.8 ) ;
		}
	Vgnd /= (double)Navg ;
	*Vground2 = Vgnd ;
	if ( qmode != 4 )
		{
		sprintf( txtbuf, ">> Rx-Test card : averaged GND = %+ 7.3fV", Vgnd ) ;
		ScrLine( 10, 1 ) ;
		}

	if ( !hwerr )
		{
    	/* ADC calibration, average Vref */
		/* select Vref input */
		WriteA( 0x0C, 0x25 ) ;
		/* wait 55ms for the input voltage to stabilize */
		Delay55ms( 0 ) ;

		Vadc = 0.0 ;
		for ( nx = Navg ; nx > 0 ; nx-- )
			{
			hwerr = ReadADC1602( &adcdata ) ;
			if ( hwerr )
				nx = 0 ;
			Vadc += 0.2 * 2.0 * ( (double)adcdata / 3276.8 ) ;
        	}
		}
	Vadc /= (double)Navg ;
	Vgain = 2.5 / ( Vadc - Vgnd ) ;
	*Vgain2 = Vgain ;
	if ( qmode != 4 )
		{
		sprintf( txtbuf, ">> Rx-Test card : averaged Vref = %+ 7.3fV  ", Vadc ) ;
		ScrLine( 11, 1 ) ;
		sprintf( txtbuf, ">> Rx-Test card : averaged Gain = %+ 7.3f   ", Vgain ) ;
		ScrLine( 12, 1 ) ;
		}
	}

if ( ( qmode & 1 ) && ( !hwerr ) )
	{
	/* retrieve Vgnd and ADCgain from the calling environment ( main() ? ) */
	Vgnd = *Vground2 ;
	Vgain = *Vgain2 ;

	/* Check supply voltages, +12V first */
	WriteA( 0x0C, 0x26 ) ;
	Delay55ms( 0 ) ;
	hwerr = ReadADC1602( &adcdata ) ;
	if ( hwerr )
		{
		sprintf( txtbuf, ">> ERROR : ADC reading problems !   hwerr = 0x%02X\n\n", hwerr ) ;
		ErrorLine( 1, 1 ) ;
		}

	Vadc = Vgain * 0.19 * 10.0 * ( (double)adcdata / 3276.8 ) - Vgnd ;

	sprintf( txtbuf, ">> Rx-Test card : +12V = %+ 7.3fV", Vadc ) ;
	ScrLine( 14, 1 ) ;

	/* Check -12V supply voltage */
	WriteA( 0x0C, 0x27 ) ;
	Delay55ms( 0 ) ;
	hwerr = ReadADC1602( &adcdata ) ;
	if ( hwerr )
		{
		sprintf( txtbuf, ">> ERROR : ADC reading problems !   hwerr = 0x%02X\n\n", hwerr ) ;
		ErrorLine( 1, 1 ) ;
		}

	Vadc = Vgain * 0.188 * 10.0 * ( (double)adcdata / 3276.8 ) - Vgnd ;
    sprintf( txtbuf, ">> Rx-Test card : -12V = %+ 7.3fV", Vadc ) ;
	ScrLine( 15, 1 ) ;
	}

if ( ( ( qmode == 10 ) || ( qmode == 0 ) ) && !hwerr )
	{
	if ( qmode == 10 )
		{
		_settextposition( 15, 1 ) ;
		_outtext( ">> Testing 10 Ohm reference ; press any key to run" ) ;
		}

	flag_kb = 0 ;
	while( !flag_kb )
		{
		/* set GND channel */
		WriteA( 0x0C, 0x24 ) ;
		flag_kb = Delay55ms( 0 ) ;

	    /* ADC calibration, average Vgnd */
		Vgnd = 0.0 ;
		for ( nx = Navg ; nx > 0 ; nx-- )
			{
			hwerr = ReadADC1602( &adcdata ) ;
			if ( hwerr )
				nx = 0 ;
			Vgnd += 0.2 * ( (double)adcdata / 3276.8 ) ;
			}
		Vgnd /= (double)Navg ;


		/* set Vref channel */
		WriteA( 0x0C, 0x25 ) ;
		flag_kb = Delay55ms( 0 ) ;

		Vadc = 0.0 ;
		for ( nx = Navg ; nx > 0 ; nx-- )
			{
			hwerr = ReadADC1602( &adcdata ) ;
			if ( hwerr )
				nx = 0 ;
			Vadc += 0.2 * 2.0 * ( (double)adcdata / 3276.8 ) ;
			}
		Vadc /= (double)Navg ;
		Vgain = 2.5 / ( Vadc - Vgnd ) ;


		/* set 10 Ohm resistor test */
		WriteA( 0x0C, 0x23 ) ;
		flag_kb = Delay55ms( 0 ) ;

		Vadc = 0.0 ;
		for ( nx = Navg ; nx > 0 ; nx-- )
			{
			hwerr = ReadADC1602( &adcdata ) ;
			if ( hwerr )
				nx = 0 ;
			Vadc += 0.2 * ( (double)adcdata / 3276.8 ) / 0.1 ;
			}
		Vadc /= (double)Navg ;
		Vadc = Vgain * ( Vadc - Vgnd ) - 0.059 ;

		if ( qmode == 10 )
			{
			sprintf( txtbuf, ">> Reference 10 Ohm resistor = %+ 7.3f Ohm     ", Vadc ) ;
			ScrLine( 16, 1 ) ;
			}
		else
			{
			sprintf( txtbuf, ">> Reference 10 Ohm resistor = %+ 7.3f Ohm     ", Vadc ) ;
			ScrLine( 13, 1 ) ;

			/* exit 'while...' loop immediately */
			flag_kb = 77 ;
			}
		}
	if ( flag_kb )
		ClearKbBuf() ;
	}

/* set GND channel */
WriteA( 0x0C, 0x24 ) ;
Delay55ms( 0 ) ;

return hwerr ;
}


/*  */
int MiliOhmRun( int pp1num, int connum, int mode )
{
int hwerr, nx, flag_kb, ferr, beepflag, testflag ;
int chnum, chset, Navg, adcdata, rtol ;
double Rxx, Vgnd, Vgain ;
char filename[13] = { "TP1D0000.TST" } ;
char filenext[5] = { ".TST" } ;
char fnroot[5] = { "TP1D" } ;
FILE *fhp ;
char d_str[9], t_str[9] ;

char *linenames[18][8] = { {"HV-Bias "},
						   {"HV-Retrn"},
 						   {"Temp-2  "},
						   {"VccSense"},
						   {"AgndSens"},
						   {"VddSense"},
						   {"DgndSens"},
						   {"Reset   "},
						   {"Temp-1  "},
						   {"Select  "},
						   {"PIN-Bias"},
						   {"ILED-b  "},
						   {"ILED-a  "},
						   {"Vdd-Supp"},
						   {"Vcc-Supp"},
						   {"Agnd    "},
						   {"Dgnd    "},
						   {"Shield  "} } ;

float Rmax[18] = {	0.500, /* HV-Bias   */
					0.500, /* HV-Return */
					0.500, /* Temp-2    */
					0.500, /* VccSense  */
					0.500, /* AgndSense */
					0.500, /* VddSense  */
					0.500, /* DgndSense */
					0.500, /* Reset     */
					0.500, /* Temp-1    */
					0.500, /* Select    */
					0.500, /* PIN-Bias  */
					0.500, /* ILED-b    */
					0.500, /* ILED-a    */
					0.100, /* VddSupply */
					0.100, /* VccSupply */
					0.100, /* Agnd      */
					0.100, /* Dgnd      */
					0.100  /* Shield    */  } ;

hwerr = 0 ;
flag_kb = 0 ;
Navg = 256 ;
fhp = '\0' ;
beepflag = 0 ;
testflag = 1 ;
fnroot[3] = flag_type ;

/* get calibration data */
hwerr = MiliOhmInit( 4, &Vgnd, &Vgain ) ;

/* test 10 Ohm reference */
/* hwerr = MiliOhmInit( 10, &Vgnd, &Vgain ) ; */

if ( mode < 0 )
	{
	/* write pp1num to the last four filename characters */
	sprintf( filename, "%s%04d%s", fnroot, pp1num, filenext ) ;

	/* open the file to which the test results are going to be written */
	fhp = fopen( filename, "a+t" ) ;
	if ( fhp == '\0' )
		{
		sprintf( txtbuf, ">> ERROR : File \"%s\" could not be opened !", filename ) ;
		ErrorLine( 1, 1 ) ;
		}
	}

if ( mode < 0 )
	{
	sprintf( txtbuf, "Rx-Test : board = %04d : socket = %1d ", pp1num, connum ) ;
	ScrLine( 4, 1 ) ;

	if ( fhp != '\0' )
		{
		ferr = fprintf( fhp, "\n%s\n", txtbuf ) ;
		ferr = fflush( fhp ) ;
		if ( ferr == EOF )
			{
			ErrorFile( 2 ) ;
			}
		_strdate( d_str ) ;
		_strtime( t_str ) ;
		sprintf( txtbuf, "-------------------- %s - %s -------", d_str, t_str ) ;
		ferr = fprintf( fhp, "%s\n", txtbuf ) ;
		ferr = fflush( fhp ) ;
		if ( ferr == EOF )
			{
			ErrorFile( 2 ) ;
			}
		}

	sprintf( txtbuf, "Line  Name       Rmax [Ohm]    Rx [Ohm]     Tol" ) ;
	ScrLine( 5, 1 ) ;

	if ( fhp != '\0' )
		{
		ferr = fprintf( fhp, "%s\n", txtbuf ) ;
		ferr = fflush( fhp ) ;
		if ( ferr == EOF )
			{
			ErrorFile( 2 ) ;
			}
		}

	for ( chnum = 0 ; chnum < 18 ; chnum ++ )
		{
		/* On PP1B, ILED-b is shorted to ILEDa, therefore skip line 12, ch.11 */
		if ( ( flag_type == 'B' ) && ( chnum == 11 ) )
			chnum++ ;

		/* On PP1F, line 3, ch.2 - Temp2 is shorted to line 12, ch.11 - ILEDb */
		/* therefore skip ch.11 */
		if ( ( flag_type == 'F' ) && ( chnum == 11 ) )
			chnum++ ;

		/* On TPP1F_D9, line 3 (ch.2 - Temp2) and line 10 (ch.9 - SELECT) */
		/* are selected by a SPDT switch - the test must be done for SELECT */
		/* therefore skip ch.2 */
		/* Also, ILED-b is disconnected, therefore skip ch.11 */
		if ( ( flag_type == 'D' ) && ( chnum == 2 ) )
			chnum++ ;
		if ( ( flag_type == 'D' ) && ( chnum == 11 ) )
			chnum++ ;

		/* set channel */
		if ( chnum < 8 )
			chset = 0x08 | ( 0x07 & chnum ) ;
		else if ( chnum < 16 )
			chset = 0x10 | ( 0x07 & chnum ) ;
		else
			chset = 0x20 | ( 0x07 & chnum ) ;

		WriteA( 0x0C, chset ) ;
		flag_kb = Delay55ms( 0 ) ;

		Rxx = 0.0 ;
		for ( nx = Navg ; nx > 0 ; nx-- )
			{
			hwerr = ReadADC1602( &adcdata ) ;
			if ( hwerr )
				nx = 0 ;
			Rxx += 0.2 * ( (double)adcdata / 3276.8 ) / 0.1 ;
			}
		Rxx /= Navg ;

		/* check offset and lead resistance and correct it !!! */
		/* Ground lead compensation : 0.020 Ohm */
		Rxx = Vgain * ( Rxx - Vgnd ) - 0.020 ;
		rtol = (int)( 100.0 * ( -1.0 + Rxx / Rmax[chnum] ) ) ;
		if ( rtol > 9999 )
			rtol = 9999 ;

	                  /* "Line  Name       Rmax [Ohm]    Rx [Ohm]     Tol" ) ; */
	                  /* 0         1         2         3         4         5      */
	                  /* 1234567890123456789012345678901234567890123456789012345 */
		sprintf( txtbuf, " % 3d  %8s      % 5.3f    %+ 7.3f     %+ 5d ", chnum+1, *linenames[chnum], Rmax[chnum], Rxx, rtol ) ;
		ScrLine( 6+chnum, 1 ) ;

		if ( fhp != '\0' )
			{
			ferr = fprintf( fhp, "%s\n", txtbuf ) ;
			ferr = fflush( fhp ) ;
			if ( ferr == EOF )
				{
				ErrorFile( 2 ) ;
				}
			}

		if ( rtol > 0 )
			{
			beepflag = 1 ;
			if ( Rxx > 15.0 )
				{
				sprintf( txtbuf, "Warning :  Line %3d disconnected ???", chnum+1 ) ;
				ErrorLine( 0, 0 ) ;

				if ( fhp != '\0' )
					{
					ferr = fprintf( fhp, "%s\n", txtbuf ) ;
					ferr = fflush( fhp ) ;
					if ( ferr == EOF )
						{
						ErrorFile( 2 ) ;
						}
					}
				}
			else
				{
				sprintf( txtbuf, "Warning :  Line %3d out of tollerance !", chnum+1 ) ;
				ErrorLine( 0, 0 ) ;

				if ( fhp != '\0' )
					{
					ferr = fprintf( fhp, "%s\n", txtbuf ) ;
					ferr = fflush( fhp ) ;
					if ( ferr == EOF )
						{
						ErrorFile( 2 ) ;
						}
					}
				}
			}
		}
	if ( beepflag )
		printf( "\a" ) ;

	/* ***** save test flags ***** */
	if ( !ferr )
		{
		if ( flag_kb || beepflag )
			testflag = 0 ;
		ferr = SaveTestFlags( pp1num, 3, connum, testflag ) ;
		}
	}
else
	{
	chnum = mode ;
	/* set channel */
	if ( chnum < 8 )
		chset = 0x08 | ( 0x07 & chnum ) ;
	else if ( chnum < 16 )
		chset = 0x10 | ( 0x07 & chnum ) ;
	else
		chset = 0x20 | ( 0x07 & chnum ) ;

	WriteA( 0x0C, chset ) ;
	flag_kb = Delay55ms( 0 ) ;

	Rxx = 0.0 ;
	for ( nx = Navg ; nx > 0 ; nx-- )
		{
		hwerr = ReadADC1602( &adcdata ) ;
		if ( hwerr )
			nx = 0 ;
		Rxx += 0.2 * ( (double)adcdata / 3276.8 ) / 0.1 ;
		}
	Rxx /= Navg ;

	/* Ground lead compensation : 0.040 Ohm */
	Rxx = Vgain * ( Rxx - Vgnd ) - 0.040 ;

	sprintf( txtbuf, ">> Line % 3d    Rx = %+ 7.3f Ohm     ", chnum+1, Rxx ) ;
	ScrLine( 6+chnum, 1 ) ;
	}

if ( fhp != '\0' )
	{
	if ( fclose( fhp ) )
		ErrorFile( 6 ) ;
	}

/* set GND channel */
WriteA( 0x0C, 0x24 ) ;
/* flag_kb = Delay55ms( 0 ) ; */

return hwerr ;
}


/*  */
int CxTestRun( int pp1num, int connum )
{
int key ;

int hwerr, flag_kb, error_flag, ferr, testerr, testflag ;
int tol30 ;
unsigned int addrx, datar1, datar2, comp16 ;
unsigned int dtstat, cmstat, kx, ky ;
unsigned int channel, maxchnum, xcon ;

unsigned long int p_cnt ;

double cap, cap15, cap16 ;

char chx ;

unsigned int bitmask[17] = { 0x0080,
							 0x0002,
                             0x8000,
                             0x0200,
                             0x0040,
                             0x0001,
                             0x4000,
                             0x0100,
                             0x0020,
                             0x0004,
                             0x2000,
                             0x0400,
                             0x0010,
                             0x0008,
                             0x1000,
                             0x0800,
                             0x0400 } ;

/* char *linenames[17][13] = { { "HV-Bias  C13" },
 *						    { "HV-Retrn C14" },
 *						    { "Temp-2   C06" },
 *						    { "VccSense C11" },
 *						    { "AgndSens C05" },
 *						    { "VddSense C08" },
 *						    { "DgndSens C04" },
 *						    { "Reset    C07" },
 *						    { "Temp-1   C03" },
 *						    { "Select   C09" },
 *						    { "PIN-Bias C02" },
 *						    { "ILED-b   C10" },
 *						    { "ILED-a   C01" },
 *						    { "Vdd-Supp C17" },
 *						    { "Vcc-Supp C16" },
 *						    { "Agnd     C12" },
 *						    { "Dgnd     C15" } } ;
 */

char *linenames[17][8] = {  { "HV-Bias " },
						    { "HV-Retrn" },
 						    { "Temp-2  " },
						    { "VccSense" },
						    { "AgndSens" },
						    { "VddSense" },
						    { "DgndSens" },
						    { "Reset   " },
						    { "Temp-1  " },
						    { "Select  " },
						    { "PIN-Bias" },
						    { "ILED-b  " },
						    { "ILED-a  " },
						    { "Vdd-Supp" },
						    { "Vcc-Supp" },
						    { "Agnd    " },
						    { "Dgnd    " } } ;

/* WARNING : ILED-a and ILED-b are shorted together on new PP1BV6ZA pcb */

float caprefs[17] = {   15.0,
					  1000.0,
					  1000.0,
					   100.0,
					   100.0,
					   100.0,
					   100.0,
					   100.0,
					  1000.0,
					  1000.0,
					  1000.0,
					  2000.0,
					  2000.0,
					  4700.0,
					  4700.0,
					  1000.0,
					  1000.0 } ;

int ctab[17] = {	13,
					14,
					 6,
					11,
					 5,
					 8,
					 4,
					 7,
					 3,
					 9,
					 2,
					10,
					 1,
					17,
					16,
					12,
					15 	} ;

int cnn ;	/* cnn = ctab[channel] ; */
char filename[13] = { "TP1D0000.TST" } ;
char filenext[5] = { ".TST" } ;
char fnroot[5] = { "TP1D" } ;
FILE *fhp ;
char d_str[9], t_str[9] ;

/* ----- initial values ----- */
hwerr = 0 ;
key = 0 ;
chx = '\0' ;
channel = 0 ;
if ( flag_type == 'B' )
	maxchnum = 15 ;
else if ( flag_type == 'F' || flag_type == 'D' )
	maxchnum = 16 ;
else
	maxchnum = 15 ;
fnroot[3] = flag_type ;
fhp = '\0' ;
testerr = 0 ;
testflag = 1 ;

/* Delta-T Mode Parameter Definition :
 ***************************************
 *					0x00 = A+ to B+
 *					0x01 = B+ to A+
 *					0x20 = A- to B+
 *					0x21 = B+ to A-
 *					0x10 = A+ to B-
 *					0x11 = B- to A+
 *					0x30 = A- to B-
 *					0x31 = B- to A-
 ***************************************
 *					0x02 = A+ to next A+
 *					0x0A = A+ to A-
 *					0x22 = A- to next A-
 *					0x2A = A- to A+
 ***************************************
 *					0x03 = B+ to next B+
 *					0x07 = B+ to B-
 *					0x13 = B- to next B-
 *					0x17 = B- to B+
 ***************************************
 */

SetDTMode( 0x2A ) ;		/* use A+ to A- ???? inv ???? */

/* write pp1num to the last four filename characters */
sprintf( filename, "%s%04d%s", fnroot, pp1num, filenext ) ;
fhp = fopen( filename, "a+t" ) ;
if ( fhp == '\0' )
	{
	ErrorFile( 1 ) ;
	}

sprintf( txtbuf, "Cx-Test : board = %04d : socket = %1d", pp1num, connum ) ;
ScrLine( 4, 1 ) ;

if ( fhp != '\0' )
	{
	ferr = fprintf( fhp, "\n%s\n", txtbuf ) ;
	ferr = fflush( fhp ) ;
	if ( ferr == EOF )
		{
		ErrorFile( 2 ) ;
		}
	_strdate( d_str ) ;
	_strtime( t_str ) ;
	sprintf( txtbuf, "-------------------- %s - %s -------", d_str, t_str ) ;
	ferr = fprintf( fhp, "%s\n", txtbuf ) ;
	ferr = fflush( fhp ) ;
	if ( ferr == EOF )
		{
		ErrorFile( 2 ) ;
		}
	}

				/*         1         2         3         4         5
				/*1234567890123456789012345678901234567890123456789012345*/
sprintf( txtbuf, "Line   Name       Cr [nF]  Cx [nF]   Cnn  Xcon  Tol30" ) ;
ScrLine( 5, 1 ) ;

if ( fhp != '\0' )
	{
	ferr = fprintf( fhp, "%s\n", txtbuf ) ;
	ferr = fflush( fhp ) ;
	if ( ferr == EOF )
		{
		ErrorFile( 2 ) ;
		}
	}

for( channel = 0 ; channel <= maxchnum ; channel++ )
	{
	/* On PP1B, ILED-b is shorted to ILEDa, therefore skip line 12, ch.11 */
	if ( ( flag_type == 'B' ) && ( channel == 11 ) )
		channel++ ;

	/* On PP1F, line 3, ch.2 - Temp2 is shorted to line 12, ch.11 - ILEDb */
	/* therefore skip ch.12 */
	if ( ( flag_type == 'F' ) && ( channel == 11 ) )
		channel++ ;

	/* On TPP1F_D9, line 3 (ch.2 - Temp2) and line 10 (ch.9 - SELECT) */
	/* are selected by a SPDT switch - the test must be done for SELECT */
	/* therefore skip ch.2 */
	/* Also, ILED-b is disconnected, therefore skip ch.11 */
	if ( ( flag_type == 'D' ) && ( channel == 2 ) )
		channel++ ;
	if ( ( flag_type == 'D' ) && ( channel == 11 ) )
		channel++ ;

	error_flag = 0 ;
	ResetCMS() ;
	cmstat = ReadCMStat() ;
	/* printf( " >> cmstat = 0x%02X        \n", cmstat ) ; */
	while( ( cmstat & 0x10 ) && !( flag_kb = kbhit() ) )
		{
		cmstat = ReadCMStat() ;
		/* printf( "\r >> cmstat = 0x%02X        ", cmstat ) ; */
		}
	if ( flag_kb = kbhit() )
		{
		error_flag = 1 ;
		ClearKbBuf() ;
		}

	Delay55ms( 0 ) ;
	StartNewDTM() ;

	/* Forward - PP1F */
	if ( flag_type == 'F' || flag_type == 'D' )
		{
		if ( channel == 13 )
			{
			/* Vdd has 4.7uF to DGND, DGND must be shorted to SYSGND */
			StartCMS( 0x80 | channel ) ; /* bit 7 High shorts ch.16 (Dgnd) to GND */
			}
		else if ( channel == 14 )
			{
			/* Vcc has 4.7uF to AGND, AGND must be shorted to SYSGND */
			StartCMS( 0x80 | channel ) ; /* bit 7 High shorts ch.15 (Agnd) to GND */
			}
		else if ( channel == 15 )
			{
			/* AGND floats on 1uF cap. and $.7uF to Vcc, Vcc must be shorted to SYSGND */
			StartCMS( 0x40 | channel ) ; /* bit 6 High shorts ch.14 (Vcc+) to GND */
			}
		else if ( channel == 16 )
			{
			/* PP1F has DGND floating on 1uF cap. and 4.7uF to Vdd */
			/* Vdd must be shorted to SYSGND */
			/* ch.16 DGND is switched to line 12, ch.11 [ILEDb], code 0x0B */
			StartCMS( 0x40 | 0x0B ) ; /* bit 6 High shorts ch.13 (Vdd+) to GND */
			}
		else
			StartCMS( channel ) ;
		}

	/* Barrel - PP1B */
	if ( flag_type == 'B' )
		{
		if ( channel == 14 )
			{
			/* Vcc has 4.7uF to AGND, AGND must be shorted to SYSGND */
			StartCMS( 0x80 | channel ) ; /* bit 7 High shorts ch.15 (Agnd) to GND */
			}
		else if ( channel == 15 )
			{
			/* AGND floats on 1uF cap. and $.7uF to Vcc, Vcc must be shorted to SYSGND */
			StartCMS( 0x40 | channel ) ; /* bit 6 High shorts ch.14 (Vcc+) to GND */
			}
		else
			StartCMS( channel ) ;
		}

	dtstat = ReadDTStat() ;

	if ( dtstat & 0x40 )
		{
		ReadDTCounter( &p_cnt ) ;
		comp16 = ReadCMS() ;
		/* printf( "\tDelta_t = % 8lu  \tcomp16 = 0x%04X   \n", p_cnt, comp16 ) ; */
		}
	else
		{
		/* test for OverRange bit-7 and DataReady bit-5 */
		_settextposition( 25, 1 ) ;
		while( ( !( dtstat & 0xC0 ) ) && ( !( flag_kb = kbhit() ) ) )
			{
			dtstat = 0xF0 & ReadDTStat() ;
			if ( dtstat == 0x40 )
				{
				ReadDTCounter( &p_cnt ) ;
				comp16 = ReadCMS() ;
				/* check operation, do not show normally */
				/* sprintf( txtbuf, "\tDelta_t = % 8lu  \tcomp16 = 0x%04X   \n", p_cnt, comp16 ) ; */
				/* ScrLine( 6+channel, 1 ) ; */
				}
			else if ( dtstat == 0x10 )
				{
				sprintf( txtbuf, ">> Delta-T Status : W4T " ) ;
				ScrLine( 6+channel, 1 ) ;
				}
			else if ( dtstat == 0x20 )
				{
				/* normal operation !!! do not show !!! */
				/* sprintf( txtbuf, ">> Delta-T Status : CNT " ) ; */
				/* ScrLine( 6+channel, 1 ) ; */
				}
			else if ( dtstat & 0x80 )
				{
				ResetDT() ;
				sprintf( txtbuf, ">> Delta-T Over Range !" ) ;
				ScrLine( 6+channel, 1 ) ;
				printf( "\a" ) ;
				error_flag = 2 ;
				WaitKey( 1 ) ;
				}
			else
				{
				dtstat = 0xF0 & ReadDTStat() ;
				if ( ( !dtstat ) || ( ( dtstat & 0x10 ) & ( dtstat & 0x20 ) ) || ( ( dtstat & 0x20 ) & ( dtstat & 0x40 ) ) || ( ( dtstat & 0x10 ) & ( dtstat & 0x40 ) ) )
					{
   	                ResetDT() ;
					sprintf( txtbuf, ">> Bad Delta-T Status : 0x%02X ", dtstat ) ;
					ScrLine( 6+channel, 1 ) ;
					printf( "\a" ) ;
					WaitKey( 1 ) ;
					}
				}
			}
		}

	if ( flag_kb = kbhit() )
		{
		error_flag = 1 ;
		ClearKbBuf() ;
		}

	/* ***** calculate cap from counter value, resolution 100ns ***** */
	/*    Vref = 3V, Vsup = 5V, R=220kohm					          */
	/*    Vref = Vsup * ( 1 - exp( -t/RC ) )				          */
	/*    ==> 1 / ( 1 - Vref/Vsup ) = exp( t/RC )		        	  */
	/*    ==> t = RC * log( 1 / ( 1 - Vref/Vsup ) )		              */
	/*    ==> C = t / R * log( 1 / ( 1 - Vref/Vsup ) )		          */

	/* ***** display cap in [nF] ***** */
	/* cap = 1e+9 * 1e-7 * ( (double)p_cnt ) / ( 2.2e+5 * log( 2.5 )  ) ; */
	cap = 1e+2 * ( (double)p_cnt ) / ( 2.2e+5 * log( 2.509 ) ) ;
	cap *= 0.96 ;
	if ( ( channel == 1 ) || ( channel == 8 ) || ( channel == 9 ) || ( channel == 10 ) || ( channel == 11 ) )
		cap *= 0.97 ;

	if ( ( flag_type == 'F' || flag_type == 'D' ) && ( channel == 13 ) )
		{
		cap = cap * 0.923 ;
		cap16 = cap ;
		}
	if ( channel == 14 )
		{
		cap = cap * 0.925 ;
		cap15 = cap ;
		}
	if ( channel == 15 )
		{
		/* cap = cap - cap15 ; */
		cap = 0.52 * ( cap - cap15 ) ;
		}
	if ( channel == 16 )
		{
		/* cap = ( cap - cap16 ) ; */
		cap = 0.67 * ( cap - cap16 ) ;
		}
	if ( cap < 0.0 )
		cap = 0.0 ;

	xcon = comp16 & ( ~bitmask[ channel ] ) ;
	for( kx = 0 ; kx <= maxchnum ; kx++ )
		{
		if ( xcon == bitmask[kx] )
			{
			xcon = kx ;
			kx = 17 ; /* show only the first error, jump the rest */
			}
		}
	if ( xcon )
		xcon++ ;	/* increment to show the same number as channel+1 */

	if ( error_flag == 2 )
		{
		cap = 0.0 ;
		xcon = -1 ;
		}
	if ( ( flag_type == 'F' || flag_type == 'D' ) && ( ( channel == 11 ) || ( channel == 12 ) ) )
		tol30 = (int)( 100.0 * ( -1.0 + cap / ( 0.5 * caprefs[channel] ) ) ) ;
	else
		tol30 = (int)( 100.0 * ( -1.0 + cap / caprefs[channel] ) ) ;
	cnn = ctab[channel] ;
	if ( ( flag_type == 'F' || flag_type == 'D' ) && ( ( channel == 11 ) || ( channel == 12 ) ) )
		sprintf( txtbuf, "% 3d    %8s   % 5.0f   % 8.2f   C%02d  % 3d   %+ 5d", channel+1, *linenames[channel], 0.5 * caprefs[channel], cap, cnn, xcon, tol30 ) ;
	else
		sprintf( txtbuf, "% 3d    %8s   % 5.0f   % 8.2f   C%02d  % 3d   %+ 5d", channel+1, *linenames[channel], caprefs[channel], cap, cnn, xcon, tol30 ) ;
	ScrLine( 6+channel, 1 ) ;

	if ( fhp != '\0' )
		{
		ferr = fprintf( fhp, "%s\n", txtbuf ) ;
		ferr = fflush( fhp ) ;
		if ( ferr == EOF )
			{
			sprintf( txtbuf, "FILE ERROR : Could not write to \"%s\" !", filename ) ;
			ErrorLine( 1, 1 ) ;
			}
		}

	if ( abs( tol30 ) > 30 )
		{
		if ( !xcon )
			{
			if ( abs( tol30 ) > 90 )
				{
				testerr = 1 ;
				sprintf( txtbuf, " >> Warning : Line %3d disconnected ???   ", channel+1 ) ;
				ErrorLine( 0, 0 ) ; /* do not wait for key !!! */

				if ( fhp != '\0' )
					{
					ferr = fprintf( fhp, "%s\n", txtbuf ) ;
					ferr = fflush( fhp ) ;
					if ( ferr == EOF )
						{
						ErrorFile( 2 ) ;
						}
					}
				}
			else
				{
				testerr = 1 ;
				sprintf( txtbuf, " >> WARNING : Line %2d : Cx out of tollerance !   ", channel+1 ) ;
				ErrorLine( 0, 0 ) ; /* do not wait key */

				if ( fhp != '\0' )
					{
					ferr = fprintf( fhp, "%s\n", txtbuf ) ;
					ferr = fflush( fhp ) ;
					if ( ferr == EOF )
						{
						ErrorFile( 2 ) ;
						}
					}
				}
			}
		else
			{
			testerr = 1 ;
			if ( xcon == -1 )
				sprintf( txtbuf, " >> Possible short to GROUND-PLANE !     " ) ;
			else
				sprintf( txtbuf, " >> Possible short with line %2d !       ", xcon ) ;
			ErrorLine( 0, 0 ) ; /* beep at the end, don't wait for key !!! */

        	if ( fhp != '\0' )
				{
				ferr = fprintf( fhp, "%s\n", txtbuf ) ;
				ferr = fflush( fhp ) ;
				if ( ferr == EOF )
					{
					ErrorFile( 2 ) ;
					}
				}
			}
		}
	if ( error_flag )
		testerr = 1 ;

	if ( error_flag > 1 )
		{
		/* DON'T exit 'for()' loop */
		/* channel = 16 ; */
		/* clear error_flag cnd continue test */
		error_flag = 0 ;
		}

	if ( error_flag == 1 )
		channel = 16 ;
	/* key : user break request : exit 'for()' loop */
	}


ResetCMS() ;

if ( fhp != '\0' )
	{
	if ( fclose( fhp ) == EOF )
		{
		ErrorFile( 6 ) ;
		ferr = 1 ;
		}
	}

/* ***** save test flags ***** */
if ( !ferr )
	{
	if ( flag_kb || testerr )
		testflag = 0 ;
	ferr = SaveTestFlags( pp1num, 1, connum, testflag ) ;
	}

hwerr = error_flag ;

return hwerr ;
}


/*  */
int GetBoardNum( void )
{
int bnumout, pp1num ;
int flag_err ;
int kin, kx, nx, intchx, intchy ;
int xpos, cpos, maxpos, insmode ;

char chx ;
char pp1numstr[4] ;
/* char txtbuf[80] ; */

kin = 0 ;
insmode = 1 ;

/* refresh the pp1num from lastpp1 (global variable) */
pp1num = lastpp1 ;

/* return -1 if no number was entered */
bnumout = -1 ;


sprintf( txtbuf, "Enter the PCB serial number (last 4 digits) : " ) ;
ScrLine( 11, 1 ) ;

/* clear any spurious content in the editing string */
for( kx = 0 ; kx <= 4 ; kx++ )
	pp1numstr[kx] = '\0' ;

/* convert the input value of pp1num to a string, radix = 10 */
/* *itoa( pp1num, pp1numstr, 10 ) ; */
sprintf( pp1numstr, "%04d", pp1num ) ;

kin = 0 ;
xpos = 4 ;
cpos = 5 ;

while ( kin < 4 )
	{
	/* find the max. cursor position */
	/* (can be max string length + 1, but no entry allowed) */
	kx = 0 ;
	while( ( pp1numstr[kx] != '\0' ) && ( kx < 4 ) )
		kx++ ;
	if ( kx >= 4 )
		kx = 4 ;
	maxpos = kx ;
	/* check and set limits */
	if ( cpos < 1 )
		cpos = 1 ;
	if ( cpos > 5 )
		cpos = 5 ;
	if ( cpos > maxpos + 1 )
		cpos = maxpos + 1 ;
	xpos = cpos - 1 ;
	pp1numstr[kx] = '\0' ;

	/* show/refresh the input value being edited */
	_settextposition( 12, 1 ) ;
	_outtext( pp1numstr ) ;

	/* clear possible spurious entries */
	_settextposition( 12, maxpos+1 ) ;
	_outtext( "     " ) ;

	/* set the cursor behind the last character */
	_settextposition( 12, cpos ) ;

	/* get the character or command from keyboard */
	intchx = getche() ;

	switch( intchx )
		{
		/* case 3592 : */ /* BACK-SPACE */
		case 8 : /* BACK-SPACE */
			{
			cpos = cpos - 1 ;
			if ( cpos < 1 )
				cpos = 1 ;
			xpos = cpos - 1 ;
			/* delete the character beforethe cursor and */
			/* shift left the remaining ones to the left */
			for ( kx = xpos ; kx < maxpos ; kx++ )
				pp1numstr[kx] = pp1numstr[kx+1] ;
			pp1numstr[maxpos-1] = '\0' ;
			break ;
			}
		case 27 : /* ESC */
			{
			/* leave string unedited */
			/* but display a warning */
			break ;
			}
		/* case 7181 : */ /* ENTER */
		case 13 : /* ENTER */
			{
			/* end the string with NULL */
			/* and exit editing */
			if ( maxpos < 4 )
				{
				sprintf( txtbuf, "WARNING : Need 4 numbers !" ) ;
				ErrorLine( 1, 1 ) ;
				_settextposition( 12, maxpos ) ;
				}
			else
				{
				pp1num = atoi( pp1numstr ) ;
				bnumout = pp1num ;
				kin = 4 ;
				}
			break ;
			}
		}

	if ( ( intchx == 0 ) || ( intchx == 0xE0 ) )
		{
		/* cursor movement key was pressed, get */
		/* the actual key-code by reading again */

		intchy = getche() ;

		switch( intchy )
			{
			/* case 19712 : */ /* RIGHT */
			case 77 : /* RIGHT */
				{
				cpos++ ;
				if ( cpos > 5 )
					cpos = 5 ;
				xpos = cpos - 1 ;
				break ;
				}
			/* case 19200 : */ /* LEFT */
			case 75 : /* LEFT */
				{
				cpos-- ;
				if ( cpos < 1 )
					cpos = 1 ;
				xpos = cpos - 1 ;
				break ;
				}
			/* case 20992 : */ /* INS */
			case 82 : /* INS */
				{
				/* toggle Insert/Overwrite mode */
				insmode = 1 - insmode ;

				/* and change the cursor accordingly */
				if ( !insmode )
					_settextcursor( 0x0007 ) ;
				else
					_settextcursor( 0x0707 ) ;
				break ;
				}
			/* case 21248 : */ /* DEL */
			case 83 : /* DEL */
				{
				if ( cpos == maxpos + 1 )
					pp1numstr[maxpos] = '\0' ;
				if ( cpos <= maxpos )
					{
					/* delete the character poined to and */
					/* shift left the remaining ones */
					for ( kx = xpos ; kx < maxpos ; kx++ )
						pp1numstr[kx] = pp1numstr[kx+1] ;
					pp1numstr[maxpos] = '\0' ;
					}
				break ;
				}
			/* case 18176 : */ /* HOME */
			case 71 : /* HOME */
				{
				/* set pointer and cursor to the */
				/* beginning of the string */
				cpos = 1 ;
				xpos = 0 ;
				break ;
				}
			/* case 20224 : */ /* END */
			case 79 : /* END */
				{
				/* set the pointer and the cursor */
				/* behind the last character of the string */
				cpos = maxpos + 1 ;
				xpos = cpos - 1 ;
				break ;
				}
			default    : /* any other key */
				{
				/* ignore other keys */
				break ;
				}
			}
		}
	else
		{
		/* check the character read and                    */
		/* put it into the string at the current position */
		/* if it is a number, otherwise display a warning */
		if ( intchx < 0x20 )
			{
			/* control-code, ignore it */
			}
		else
			{
			if ( ( (char)intchx < '0' ) || ( (char)intchx > '9' ) )
				{
				sprintf( txtbuf, ">> ERROR : Use numbers only, please !" ) ;
				ErrorLine( 1, 1 ) ;
				}
			else
				{
				if ( ( cpos < 5 ) && ( xpos < 4 ) )
					{
					if ( insmode )
						{
						for ( kx = maxpos ; kx >= cpos ; kx-- )
							pp1numstr[kx] = pp1numstr[kx-1] ;
						}
					pp1numstr[xpos] = (char)intchx ;
					xpos++ ;
					cpos++ ;
					if ( ( xpos > 4 ) || ( cpos > 5 ) )
						{
						xpos = 4 ;
						cpos = 5 ;
						}
					}
				}
			}
		}
	}

if ( _gettextcursor() != 0x0707 )
	{
	insmode = 0 ;
	_settextcursor( 0x0707 ) ;
	}

return bnumout ;
}


/*  */
int GetSockNum( void )
{
int connum, maxnum ;
char chx ;

if ( flag_type == 'F' )
	maxnum = 3 ;
else if ( flag_type == 'D' )
	maxnum = 9 ; /* use MAXSOCKNUM */
else
	maxnum = 6 ;

connum = 0 ;
while ( connum < 1 || connum > maxnum )
	{
	ClearLines( 13, 13 ) ;
	sprintf( txtbuf, "Enter the connector number [ 1 - %d ] : ", maxnum ) ;
	ScrLine( 13, 1 ) ;

	/* kin = cscanf( "%d", &connum ) ; */
	chx = getche() ;
	if ( ( chx > '0' ) && ( chx <= '9' ) )
		connum = chx - 0x30 ;
	}

return connum ;
}


/*  */
char GetPCBType( void )
{
char chx, chbf ;
int x, swstat, match ;

x = 0 ;
match = 0 ;
while ( !x )
	{
	ClearLines( 13, 13 ) ;
	sprintf( txtbuf, "Enter the PCB/Type [ B, F or D ] : " ) ;
	ScrLine( 13, 1 ) ;

	chx = getche() ;
	if ( ( chx != 'F') && ( chx != 'f' ) && ( chx != 'B' ) && ( chx != 'b' ) && ( chx != 'D' ) && ( chx != 'd' ) )
		{
		ClearKbBuf() ;
		printf( "\a" ) ;
		}
	else
		x = 1 ;
	}

ClearLines( 14, 16 ) ;
if ( chx == 'b' )
	chx = 'B' ;
if ( chx == 'f' )
	chx = 'F' ;
if ( chx == 'd' )
	chx = 'D' ;

while ( !match )
	{
    swstat = 0x40 & ReadCMStat() ;
	sprintf( txtbuf, "swstat = 0x%02X", swstat ) ;
	ScrLine( 14, 1 ) ;
	if ( ( chx == 'B' ) && ( swstat ) )
		{
		chbf = 'B' ;
		match = 1 ;
		}
	else if ( ( chx == 'F' || chx == 'D' ) && ( !swstat ) )
		{
		chbf = 'F' ;
		match = 1 ;
		}
	else
		match = 0 ;
	if ( !match )
		{
		sprintf( txtbuf, "Please, set the B/F-switch on the RC-tets module to %c ", chbf ) ;
		ScrLine( 15, 1 ) ;
		sprintf( txtbuf, "and press the ENTER key !" ) ;
		ScrLine( 16, 1 ) ;
		printf( "\a" ) ;

		/* wait for ENTER key */
		while( WaitKey( 1 ) != 0x1C0D )
			{ ; }
		ClearLines( 13, 16 ) ;
		}
	}

ClearLines( 13, 16 ) ;

return chx ;
}


/*  */
int Exq ( int key )
{
unsigned hwpres ;
int hwerr, ferr, flag_quit ;
int keydown, kesc, fileok, linenum ;
int kx, nx, intchx, intchy ;
int insmode, newlist ;

int pp1num, connum, kin ;
int dtstat, cmstat ;

double Vground1, Vgain1, Vground2, Vgain2, Vadcx ;
double HVavg, Ilkgavg, Ilkravg ;

char chx ;
/* char txtbuf[80] ; */
char filename[13] ;

hwerr = 0 ;
hwpres = 0xFFFF ;
flag_quit = 0 ;
insmode = 1 ; /* character insert-mode = 1, overwrite-mode == 0 */

Vground1 = 0.0 ;
Vground2 = 0.0 ;
Vgain1 = 1.0 ;
Vgain2 = 1.0 ;

sprintf( filename, "%s", "" ) ;

if ( key > 0 )
	{
	ClearLines( 24, 25 ) ;
	switch ( key )
		{
		case 15104 : /* F1 */
			{
			/* Hardware Test */

			ClearKbBuf() ;

			/* reset the Delta-T card */
			ResetDT() ;

			/* reset the Cx-Test card */
			ResetCMS() ;

			/* Check Delta-T card presence */
			SetDTMode( 0x05 ) ;
		    dtstat = ReadDTStat() ;
			if ( ( dtstat & 0x0F ) != 0x05 )
				hwpres &= 0xFFFE ;

			SetDTMode( 0x0A ) ;
		    dtstat = ReadDTStat() ;
			if ( ( dtstat & 0x0F ) != 0x0A )
				hwpres &= 0xFFFE ;

			SetDTMode( 0x2A ) ;
			ResetDT() ;

			if ( !( hwpres & 0x01 ) )
				{
				sprintf( txtbuf, ">> ERROR : Delta-T card not found at address 0x00-0x03 !" ) ;
				ErrorLine( 1, 1 ) ;
				}

			/* Check Capacitor test card presence */
			WriteA( 0x00, 0x05 ) ;
			cmstat = ReadCMStat() ;
			if ( ( cmstat & 0x0F ) != 0x05 )
				hwpres &= 0xFFFD ;

			WriteA( 0x00, 0x0A ) ;
			cmstat = ReadCMStat() ;
			if ( ( cmstat & 0x0F ) != 0x0A )
				hwpres &= 0xFFFD ;

			ResetCMS() ;

			if ( !( hwpres & 0x02 ) )
				{
				sprintf( txtbuf, ">> ERROR : Cx-Test card not found at address 0x04-0x07 !" ) ;
				ErrorLine( 1, 1 ) ;
				}
			/* Run a continuous Cap test on lines 1 & 2 ???? */

			/* Check High-Voltage card presence */
			/* Check ADC conversion, power supply and Vref levels */
			hwerr = HVTestInit( 2, &Vground1, &Vgain1 ) ;
			if ( hwerr )
				hwpres &= 0xFFFB ;

			if ( !( hwpres & 0x04 ) )
				{
				sprintf( txtbuf, ">> ERROR : HV-Test card not found at address 0x08-0x0B !" ) ;
				ErrorLine( 1, 1 ) ;
				}
			/* Run a continuous HV test ??? */

			/* Check Resistance test card presence */
			/* Check ADC conversion, power supply and Vref levels */
			hwerr = MiliOhmInit( 2, &Vground2, &Vgain2 ) ;
			if ( hwerr )
				hwpres &= 0xFFF7 ;

			if ( !( hwpres & 0x08 ) )
				{
				sprintf( txtbuf, ">> ERROR : Rx-Test card not found at address 0x0C-0x0F !" ) ;
				ErrorLine( 1, 1 ) ;
				}

			if ( hwpres == 0xFFFF )
				{
				hwerr = HVTestInit( 0, &Vground1, &Vgain1 ) ;
				if ( hwerr )
					{
					sprintf( txtbuf, ">> ERROR : HV-Test card not initiallized !" ) ;
					ErrorLine( 1, 1 ) ;
					}

				hwerr = HVTestRun( -1, -1, 0 ) ;
				if ( hwerr )
					{
					sprintf( txtbuf, ">> ERROR : HV-Test card not running !" ) ;
					ErrorLine( 1, 1 ) ;
					}

				hwerr = MiliOhmInit( 1, &Vground2, &Vgain2 ) ;
				if ( hwerr )
					{
					sprintf( txtbuf, ">> ERROR : Rx-Test card not initiallized !" ) ;
					ErrorLine( 1, 1 ) ;
					}

				hwerr = MiliOhmInit( 0, &Vground2, &Vgain2 ) ;
				if ( hwerr )
					{
					sprintf( txtbuf, ">> ERROR : Rx-Test card not runnung !" ) ;
					ErrorLine( 1, 1 ) ;
					}
				}

			if( 0 )
				{
				while( !kbhit() )
					{
					/* Run a continuous Rx-test on line 1 and 20 ??? */
					hwerr = MiliOhmRun( pp1num, connum, 5 ) ;
					hwerr = MiliOhmRun( pp1num, connum, 0 ) ;
					}
				}
			ClearKbBuf() ;

			break ;
			}
		case 15360 : /* F2 */
			{
			/* prepare to run Cx-Test :
			 * lastpp1 is a global variable, containig
			 * the number of the last board tested ;
			 * ask for new pp1num and connum
			 */

			pp1num = GetBoardNum() ;
			if ( pp1num < 0 )
				{
				/* number was not edited, use the old one */
				pp1num = lastpp1 ;
				}
			else
				{
				/* update the global variable */
				lastpp1 = pp1num ;
				}

			/* get the connector socket number */
			connum = GetSockNum() ;

			_settextposition( 15, 1 ) ;
			_outtext( "Insert the Cx-Test plug into :" ) ;
			sprintf( txtbuf, "board %04d,   socket %1d", pp1num, connum ) ;
			ScrLine( 16, 10 ) ;
			_settextposition( 17, 1 ) ;
			_outtext( "and press ENTER ... " ) ;

			/* generate at least 1s delay to prevent immediate ENTER key */
			Delay55ms( 19 ) ;
			ClearKbBuf() ;

			while( WaitKey( 1 ) != 7181 )
				{ ; }
			ClearLines( 11, 17 ) ;

			hwerr = CxTestRun( pp1num, connum ) ;
			/* if hwerr, test was not completed */

			break ;
			}
		case 15616 : /* F3 */
			{
			/* prepare for High-Voltage test */
			pp1num = GetBoardNum() ;
			if ( pp1num < 0 )
				{
				/* number was not edited, use the old one */
				pp1num = lastpp1 ;
				}
			else
				{
				/* update the global variable */
				lastpp1 = pp1num ;
				}

			/* get the connector socket number */
			connum = GetSockNum() ;

			_settextposition( 15, 1 ) ;
			_outtext( "Insert the HV-Test plug into :" ) ;
			sprintf( txtbuf, "board %04d,   socket %1d", pp1num, connum ) ;
			ScrLine( 16, 10 ) ;
			_settextposition( 17, 1 ) ;
			_outtext( "and press ENTER ... " ) ;

			/* generate at least 1s delay to prevent immediate ENTER key */
			Delay55ms( 19 ) ;

			while( WaitKey( 1 ) != 7181 )
				{ ; }
			ClearLines( 11, 17 ) ;

			ClearKbBuf() ;
			hwerr = HVTestRun( pp1num, connum, 30 ) ;
			/* if hwerr, test was not completed !!! */
			ClearKbBuf() ;

			break ;
			}
		case 15872 : /* F4 */
			{
			/* prepare for Rx-Test run */
			pp1num = GetBoardNum() ;
			if ( pp1num < 0 )
				{
				/* number was not edited, use the old one */
				pp1num = lastpp1 ;
				}
			else
				{
				/* update the global variable */
				lastpp1 = pp1num ;
				}

			/* get the connector socket number */
			connum = GetSockNum() ;

			_settextposition( 15, 1 ) ;
			_outtext( "Insert the Rx-Test plug into :" ) ;
			sprintf( txtbuf, "board %04d,   socket %1d", pp1num, connum ) ;
			ScrLine( 16, 10 ) ;
			_settextposition( 17, 1 ) ;
			_outtext( "short all the solder pads and press ENTER ... " ) ;

			/* generate at least 1s delay to prevent immediate ENTER key */
			Delay55ms( 19 ) ;
			ClearKbBuf() ;

			while( WaitKey( 1 ) != 7181 )
				{ ; }
			ClearLines( 11, 17 ) ;

			hwerr = MiliOhmRun( pp1num, connum, -1 ) ;
			/* if hwerr, test was not completed */

			break ;
			}
		case 16128 : /* F5 */
			{
			/* file list */
			dispmenu( 1 ) ;
			key = 0 ;
			newlist = 1 ;
			while( key != 283 /*ESC*/ )
				{
				ferr = SelectFile( filename, newlist ) ;
				/* clear newlist flag for subsequent function calls */
				newlist = 0 ;

				if ( !ferr )
					{
					sprintf( txtbuf, "Selected File : %12s", filename ) ;
					ScrLine( 24, 1 ) ;
					ListFile( filename ) ;
					}
				else
					key = 283 ;
				}
			key = 0 ;
			ClearLines( 3, 25 ) ;
			dispmenu( 0 ) ;
			break ;
			}
		case 16384 : /* F6 */
			{
			lastpp1 = GetLastFile( 0 ) ;
			break ;
			}
		case 16640 : /* F7 */
			{
			lastpp1 = GetLastFile( 1 ) ;
			break ;
			}
		case 16896 : /* F8 */
			{
			flag_type = GetPCBType() ;
			break ;
			}
		case 17152 : /* F9 */
			{
			break ;
			}
		case 17408 : /* F10 */
			{
			break ;
			}
		case 19712 : /* RIGHT */
			{
			break ;
			}
		case 19200 : /* LEFT */
			{
			break ;
			}
		case 18432 : /* UP */
			{
			break ;
			}
		case 20480 : /* DOWN */
			{
			break ;
			}
		case 20992 : /* INS */
			{
			break ;
			}
		case 21248 : /* DEL */
			{
			break ;
			}
		case 18176 : /* HOME */
			{
			break ;
			}
		case 20224 : /* END */
			{
			break ;
			}
		case 18688 : /* PGUP */
			{
			break ;
			}
		case 20736 : /* PGDN */
			{
			break ;
			}
		case 283   : /* ESC */
			{
			flag_quit = 1 ;
			break ;
			}
		case 7181  : /* ENTER */
			{
			/* flag_quit = 0 ; */
			break ;
			}
		default    : /* any other key */
			break ;
		}
	}

return flag_quit ;
}


/*  */

void main( void )
{
int hwerr, flag_quit, firstrun ;
int key, kx ;

/* char txtbuf[80] ; */

key = 0 ;
firstrun = 1 ;
hwerr = 0 ;
flag_quit = 0 ;
flag_test = 0 ;

ClearKbBuf() ;

/* clear filelist */
for( kx = 0 ; kx < 22 ; kx++ )
	{
	filelist[kx][1] = '\0' ;
	}

dispmenu( 0 ) ;
if ( !flag_test )
	hwerr = InitPrnPort() ;

if ( hwerr && !flag_test )
	{
	flag_quit = 1 ;
	}

/* find the serial number of the last board tested */
lastpp1 = GetLastFile( 1 ) ;

while ( ( !flag_quit ) && ( key != 283 ) )
	{
	/* key = 283 (ESC) ==> exit program */
	ClearKbBuf() ;
	if ( firstrun )
		{
		key = 15104 ;
		/* F1 - harware test */
		firstrun = 0 ;
		}
	else
		{
		key = WaitKey( 1 ) ;
		/* clear screen before running a test */
		ClearLines( 3, 25 ) ;
		/* first define the PCB type ( B or F ) */
		if ( ( flag_type != 'B' ) && ( flag_type != 'F' ) && ( flag_type != 'D' ) )
			key = 16896 ;
		}

	flag_quit = Exq( key ) ;
	if ( key != 0 )
		{
		/* show key code for testing */
		sprintf( txtbuf, "Key Code : Hex 0x%04X,  Dec % 6d ", key, key ) ;
		ScrLine( 25, 15 ) ;
		}
	}

_clearscreen( _GCLEARSCREEN ) ;

exit ( 0 ) ;
}

/*  END  */

